/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "IConfig.h"
#include "ui_IConfig.h"
#include "ConfigMgr.h"
#include <QMessageBox>

IConfig::IConfig(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::IConfig)
{
    ui->setupUi(this);

    layout = new QHBoxLayout(this);
    servers = new IConfigServers;
    options = new IConfigOptions;
    logging = new IConfigLogging;

    layout->addWidget(servers);
    layout->addWidget(options);
    layout->addWidget(logging);
    ui->frame->setLayout(layout);

    connect(ui->toolServers, &QAbstractButton::released, [this](){ showSubDialog(SubDialogType::Servers); });
    connect(ui->toolOptions, &QAbstractButton::released, [this](){ showSubDialog(SubDialogType::Options); });
    connect(ui->toolLogging, &QAbstractButton::released, [this](){ showSubDialog(SubDialogType::Logging); });

    showSubDialog(SubDialogType::Servers);
}

IConfig::~IConfig()
{
    delete ui;
}

void IConfig::showDisconnectButton()
{
    ui->btnDisconnect->show();
}

void IConfig::hideDisconnectButton()
{
    ui->btnDisconnect->hide();
}

void IConfig::showSubDialog(IConfig::SubDialogType dlg)
{
    servers->hide();
    options->hide();
    logging->hide();
    ui->toolServers->setChecked(false);
    ui->toolOptions->setChecked(false);
    ui->toolLogging->setChecked(false);

    switch (dlg) {
    case SubDialogType::Servers:
        servers->show();
        ui->toolServers->setChecked(true);
        break;
    case SubDialogType::Options:
        options->show();
        ui->toolOptions->setChecked(true);
        break;
    case SubDialogType::Logging:
        logging->show();
        ui->toolLogging->setChecked(true);
        break;
    }
}

void IConfig::saveAll()
{
    servers->save();
    options->save();
    logging->save();

    ConfigMgr::instance().save();
}

bool IConfig::askForSave()
{
    bool serversChanged = servers->isChanged();
    bool optionsChanged = options->isChanged();
    bool loggingChanged = logging->isChanged();
    if (serversChanged || optionsChanged || loggingChanged) {
        auto btn = QMessageBox::question(this, tr("Changes made"), tr("Do you want to save the changes?"), QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel);
        if (btn == QMessageBox::Yes) {
            saveAll();
        }
        else if (btn == QMessageBox::No) {
            servers->reset();
            options->reset();
            logging->reset();
        }
        else if (btn == QMessageBox::Cancel) {
            return false;
        }
    }
    return true;
}

void IConfig::on_btnSave_clicked()
{
    saveAll();
}

void IConfig::on_btnSaveConnect_clicked()
{
    saveAll();
    ConfigMgr& conf = ConfigMgr::instance();
    if (conf.connection("Realname").isEmpty()) {
        QMessageBox::information(this, tr("Missing field"), tr("Real name must be filled."));
        return;
    }

    if (conf.connection("Username").isEmpty()) {
        QMessageBox::information(this, tr("Missing field"), tr("Username/email must be filled."));
        return;
    }

    if (conf.connection("Nickname").isEmpty()) {
        QMessageBox::information(this, tr("Missing field"), tr("Nickname must be filled"));
        return;
    }

    if (!servers->connectToNewStatus())
        emit disconnectFromServer();
    emit connectToServer(servers->connectToNewStatus());
    servers->unsetConnectToNewStatus();
    close();
}

void IConfig::on_btnDisconnect_clicked()
{
    ui->btnDisconnect->hide();
    emit disconnectFromServer();
}

void IConfig::on_btnClose_clicked()
{
    servers->unsetConnectToNewStatus();
    close();
}

void IConfig::closeEvent(QCloseEvent* evt)
{
    if (askForSave())
        evt->accept();
    else
        evt->ignore();
}
