/*
 * IdealIRC Core - Internet Relay Chat API
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "DCC.h"
#include "IRCBase.h"
#include "NATUtils/PortMapping.h"
#include "NATUtils/PublicAddress.h"
#include "Utilities.h"
#include <fmt/format.h>
#include <iostream>

using asio::ip::tcp;

struct DCCPriv
{
    DCCPriv(DCC& super_, IRCBase& ircctx_, DCC::Direction dir_, asio::io_context& ioctx_, const std::string& ip_, const std::string& port_)
        : super(super_)
        , ircctx(ircctx_)
        , direction(dir_)
        , ioctx(ioctx_)
        , ip(ip_)
        , port(port_)
        , resolver(ioctx_)
        , acceptor(ioctx_)
        , endpoint(tcp::v4(), std::stoi(port_))
        , socket(ioctx_)
        , readbuf(16384)
    {
    }

    DCC& super;
    IRCBase& ircctx;

    bool pending{ true };
    const DCC::Direction direction;
    asio::io_context& ioctx;
    std::string ip; // Only used when Direction=Target, and describes the Initiator's IP.
    std::string port;

    tcp::resolver resolver; // for outbound connection (we act as a Target)
    tcp::acceptor acceptor; // for inbound connection (we act as an Initiator)

    tcp::resolver::results_type resolverResults;
    asio::ip::tcp::endpoint endpoint;
    tcp::socket socket;

    DCC::ByteString readbuf;

    DCC::CallbackRead cbRead;
    DCC::CallbackConnected cbCon;
    DCC::CallbackDisconnected cbDiscon;

    void connected(const asio::error_code& ec)
    {
        super.onConnected();
        read(ec);
    }

    void disconnected(const asio::error_code& ec)
    {
        try {
            socket.close();
            super.onDisconnected(SystemErrorToIRCError(ec));
        }
        catch (...) {} // Might already be closed, and so we don't care about any exceptions.
    }

    void read(const asio::error_code& ec)
    {
        socket.async_receive(asio::buffer(readbuf.data(), readbuf.size()),
                             [this](const asio::error_code& ec, std::size_t rl) {
                                 if (!ec) {
                                     std::cout << "DCC read " << rl << " bytes." << std::endl;
                                     super.onRead(readbuf, rl);
                                     read(ec);
                                 }

                                 /* Operation aborted would happen during destruction and socket closing has already taken place. */
                                 else if (ec != asio::error::operation_aborted) {
                                     disconnected(ec);
                                 }
                             });
    }

    void writeFinished(const asio::error_code& ec, std::size_t bytes_transferred)
    {
        std::cout << "DCC wrote " << bytes_transferred << " bytes." << std::endl;
        if (!ec) return; // No errors.
        std::cerr << fmt::format("'- DCC write error: {} ({})", SystemErrorToIRCError(ec), ec.message()) << std::endl;
    }

    void startListener()
    {
        acceptor.open(endpoint.protocol());
        acceptor.set_option(tcp::acceptor::reuse_address(true));
        acceptor.bind(endpoint);
        acceptor.listen(1);
        acceptor.async_accept(socket, [this](const asio::error_code& ec){
            if (ec == asio::error::operation_aborted)
                return;

            if (ec)
                disconnected(ec);
            else
                connected(ec);
        });
    }
};

DCC::DCC(IRCBase& ircctx, asio::io_context& ioctx)
        : mp(std::make_unique<DCCPriv>(*this, ircctx, Direction::Initiator, ioctx, "", "0"))
{
    auto ipAddress = findOwnIpAddress();

    mp->startListener(); // also acquires a port for us
    mp->port = std::to_string(port());

    if (isPrivateIp(ipAddress) && NATPortMapping::add(port(), port())) {
        auto publicIp = NATPublicAddress::get();
        if (publicIp.has_value()) {
            ipAddress = longIpToNormal(*publicIp);
        }
        else {
            NATPortMapping::remove(port(), port());
            // Reset internal state, effectively stopping the listening.
            // This DCC instance will be destroyed and replaced when we get a response with the port to use in reverse.
            mp = std::make_unique<DCCPriv>(*this, ircctx, Direction::Initiator, ioctx, "", "0");
        }
    }

    mp->ip = ipAddress;
}

DCC::DCC(IRCBase& ircctx, asio::io_context& ioctx, const std::string& ip, const std::string& port)
    : mp(std::make_unique<DCCPriv>(*this, ircctx, Direction::Target, ioctx, ip, port))
{}

DCC::DCC(DCC&& other) noexcept
    : mp(std::move(other.mp))
{}

DCC::~DCC()
{
    NATPortMapping::remove(port(), port());
    disconnect();
}

bool DCC::isPending() const
{
    return mp->pending;
}

IRCError DCC::accept(const std::string& targetNickname)
{
    if (mp->direction == Direction::Initiator)
        return IRCError::DCC_NotATarget; // Only usable if we are a Target...

    auto acceptNormally = [this] {
        mp->resolverResults = mp->resolver.resolve(mp->ip, mp->port);

        asio::async_connect(mp->socket, mp->resolverResults,
                            [this](const asio::error_code& ec, const tcp::endpoint&) {
                                if (ec)
                                    mp->disconnected(ec);
                                else
                                    mp->connected(ec);
                            });

        mp->pending = false;
        return IRCError::NoError;
    };

    auto acceptReversed = [this, &targetNickname] {
        auto ipAddress = findOwnIpAddress();
        mp->startListener();

        if (isPrivateIp(ipAddress) && NATPortMapping::add(port(), port())) {
            auto publicIp = NATPublicAddress::get();
            if (publicIp.has_value())
                ipAddress = longIpToNormal(*publicIp);
            else
                NATPortMapping::remove(port(), port());
        }

        mp->ircctx.ctcpRequest(targetNickname, "DCC", fmt::format("CHAT chat {} {}",
                                                                  normalIpToLong(ipAddress),
                                                                  port()));

        mp->pending = false;
        return IRCError::NoError;
    };

    if (mp->port == "0")
        return acceptReversed();
    else
        return acceptNormally();
}

void DCC::disconnect()
{
    try {
        if (mp->socket.is_open()) {
            mp->socket.shutdown(asio::socket_base::shutdown_both);
            mp->socket.close();
        }
        if (mp->direction == Direction::Initiator)
            mp->acceptor.close();
    }
    catch(const asio::system_error& e) {
        std::cerr << fmt::format("DCC::disconnect() caught exception ({}): {}\n       IP={}, Port={}, Direction={}",
                                 e.code().value(), e.what(), mp->ip, mp->port, static_cast<int>(mp->direction)) << std::endl;
    }
}

DCC::Direction DCC::direction() const
{
    return mp->direction;
}

void DCC::write(const ByteString& data)
{
    mp->socket.async_send(asio::buffer(data),
        [this](const asio::error_code& ec, std::size_t bytes_transferred){
            mp->writeFinished(ec, bytes_transferred);
        });
}

IRCBase& DCC::context()
{
    return mp->ircctx;
}

const IRCBase& DCC::context() const
{
    return mp->ircctx;
}

bool DCC::isConnected() const
{
    return mp->socket.is_open();
}

uint16_t DCC::port() const
{
    if (mp->port.empty() || mp->port == "0")
        return mp->acceptor.local_endpoint().port();
    else
        return std::stoi(mp->port);
}

const std::string& DCC::ip() const
{
    return mp->ip;
}

bool DCC::isReversed() const
{
    return mp->port == "0";
}

void DCC::setCallbackRead(DCC::CallbackRead&& cb)
{
    mp->cbRead = std::move(cb);
}

void DCC::setCallbackConnected(DCC::CallbackConnected&& cb)
{
    mp->cbCon = std::move(cb);
}

void DCC::setCallbackDisconnected(DCC::CallbackDisconnected&& cb)
{
    mp->cbDiscon = std::move(cb);
}


void DCC::onRead(const DCC::ByteString& data, std::size_t length)
{
    if (mp->cbRead)
        mp->cbRead(data, length);
}

void DCC::onConnected()
{
    if (mp->cbCon)
        mp->cbCon();
}

void DCC::onDisconnected(IRCError e)
{
    if (mp->cbDiscon)
        mp->cbDiscon(e);
}

void DCC::resetCallbacks()
{
    mp->cbRead = {};
    mp->cbCon = {};
    mp->cbDiscon = {};
}

DCC::ByteString DCC::stringToByteString(const std::string& input)
{
    std::basic_string<std::byte> temp(reinterpret_cast<const std::byte*>(input.data()), input.size());
    DCC::ByteString output(temp.begin(), temp.end());
    return output;
}
