/*
 * IdealIRC Core - Internet Relay Chat API
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#ifndef DCC_H
#define DCC_H

#include "IRCError.h"
#include <asio.hpp>
#include <vector>
#include <memory>

using asio::ip::tcp;

class IRCBase;

struct DCCPriv;
class DCC
{
    friend struct IRCBasePriv;
    friend struct DCCPriv;

public:
    using ByteString = std::vector<std::byte>;
    using CallbackRead = std::function<void(const ByteString&, std::size_t)>;
    using CallbackConnected = std::function<void()>;
    using CallbackDisconnected = std::function<void(IRCError e)>;

    /*
     * Initiator acts as the server, and the target is the client
    */
    enum class Direction {
        Initiator,
        Target
    };

    virtual ~DCC();
    DCC(DCC&& other) noexcept;

    //! Constructs an Initiator (server)
    DCC(IRCBase& ircctx, asio::io_context& ioctx);

    //! Constructs a Target (client)
    DCC(IRCBase& ircctx, asio::io_context& ioctx, const std::string& ip, const std::string& port);

    [[nodiscard]] bool isPending() const;
    IRCError accept(const std::string& targetNickname);

    void disconnect();

    [[nodiscard]] Direction direction() const;

    void write(const ByteString& data);

    IRCBase& context();
    const IRCBase& context() const;

    bool isConnected() const;

    /*
     * As initiator: Port to listen on.
     * Reversed initiator: Always zero.
     * As target: Port to connect to.
    */
    uint16_t port() const;

    /*
     * As initiator: IP to listen on (published in CTCP DCC request)
     * As target or reversed initiator: IP to connect to.
    */
    const std::string& ip() const;

    bool isReversed() const;

    /*
     * Callbacks used by 'users' of this class.
    */
    void setCallbackRead(CallbackRead&& cb);
    void setCallbackConnected(CallbackConnected&& cb);
    void setCallbackDisconnected(CallbackDisconnected&& cb);

    void resetCallbacks();

    static ByteString stringToByteString(const std::string& input);

protected:
    /*
     * Use these to read data in a custom derivative of DCC.
     * If you re-implement any of these, make sure to forward the
     * calls to this base class at the end of your function.
    */
    virtual void onRead(const ByteString& data, std::size_t length);
    virtual void onConnected();
    virtual void onDisconnected(IRCError e);

private:
    std::unique_ptr<DCCPriv> mp;
};

#endif // DCC_H
