/*
 * IdealIRC Core - Internet Relay Chat API
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "IRCBasePriv.h"
#include "IRCBase.h"
#include "Utilities.h"
#include "Commands.h"

#include "IRCMember.h"
#include "IRCChannel.h"
#include "IRCMessage.h"
#include "DCC.h"

#include <fmt/format.h>

#include <iostream>
#include <stdexcept>

IRCBase::IRCBase()
    : mp(new IRCBasePriv(*this))
{
    std::copy(V3Support.begin(), V3Support.end(), std::back_inserter(mp->ircv3enabled));
}

IRCBase::IRCBase(IRCBase&& other) noexcept
    : mp(std::move(other.mp))
{}

IRCBase::~IRCBase() = default;

bool IRCBase::poll()
{
    asio::error_code ioctxerr;
    bool ioDidSomething = false;
    try {
        ioDidSomething = mp->ioctx.poll_one(ioctxerr) > 0;
    }
    catch (const std::system_error& e) {
        mp->lastError = SystemErrorToIRCError(e);
        mp->disconnectHandler();
        return false;
    }

    if (ioctxerr.value() != 0)
        std::cout << fmt::format("IOCTX err {}: {}", ioctxerr.value(), ioctxerr.message()) << std::endl;
    return ioDidSomething;
}

const std::string& IRCBase::getHostname() const
{
    return mp->hostname;
}

IRCError IRCBase::setHostname(const std::string& hostname, bool SSL)
{
    if (isOnline())
        return IRCError::CannotChangeWhenConnected;
    mp->useSSL = SSL;
    mp->hostname = hostname;
    return IRCError::NoError;
}

const std::string& IRCBase::getPort() const
{
    return mp->port;
}

IRCError IRCBase::setPort(const std::string& port)
{
    if (isOnline())
        return IRCError::CannotChangeWhenConnected;
    mp->port = port;
    return IRCError::NoError;
}

const std::string& IRCBase::getRealname() const
{
    return mp->realname;
}

IRCError IRCBase::setRealname(const std::string& realname)
{
    if (isOnline())
        return IRCError::CannotChangeWhenConnected;
    mp->realname = realname;
    return IRCError::NoError;
}

const std::string& IRCBase::getIdent() const
{
    return mp->ident;
}

IRCError IRCBase::setIdent(const std::string& ident)
{
    if (isOnline())
        return IRCError::CannotChangeWhenConnected;
    mp->ident = ident;
    return IRCError::NoError;
}

const std::string& IRCBase::getNickname() const
{
    return mp->nickname;
}

void IRCBase::setNickname(const std::string& nickname)
{
    if (mp->isConnected) {
        if (!isOnline())
            mp->nickname = nickname;
        mp->write(Command::IRC::NICK, nickname);
    }
    else
        mp->nickname = nickname;
}

const std::string& IRCBase::getPassword() const
{
    return mp->password;
}

void IRCBase::setPassword(const std::string& password)
{
    mp->password = password;
}

const std::string& IRCBase::getV3SASL() const
{
    return mp->v3SASL;
}

void IRCBase::setV3SASL(const std::string& credentials)
{
    mp->v3SASL = credentials;
}

void IRCBase::exceptSSL_SelfSigned(bool except)
{
    mp->sslExcept.selfSigned = except;
}

void IRCBase::exceptSSL_CNMismatch(bool except)
{
    mp->sslExcept.CNMismactch = except;
}

void IRCBase::exceptSSL_Expired(bool except)
{
    mp->sslExcept.expired = except;
}

void IRCBase::command(const std::string& command, const std::vector<std::string>& args, const std::string& msg)
{
    if (!mp->isConnected)
        return;

    if (msg.empty())
        mp->writeNoMsg(command, args);
    else
        mp->write(command, args, msg);
}

void IRCBase::command(const std::string& command, const std::string& msg)
{
    if (!mp->isConnected)
        return;

    mp->write(command, msg);
}

void IRCBase::raw(const std::string& data)
{
    if (!mp->isConnected)
        return;

    mp->writeNoMsg(data, {});
}

void IRCBase::ctcpRequest(const std::string& target, const std::string& command, const std::string& message)
{
    mp->ctcp(Command::IRC::PRIVMSG, target, command, message);
}

void IRCBase::ctcpResponse(const std::string& target, const std::string& command, const std::string& message)
{
    mp->ctcp(Command::IRC::NOTICE, target, command, message);
}

void IRCBase::setManualKeepalive(std::chrono::seconds freq)
{
    mp->keepaliveFreq = freq;
    if (mp->isOnline) {
        if (freq > std::chrono::seconds(0))
            mp->startKeepaliveTimer();
        else
            mp->stopKeepaliveTimer();
    }
}

std::chrono::milliseconds IRCBase::getManualKeepaliveFreq() const
{
    return mp->keepaliveFreq;
}

asio::io_context& IRCBase::getIOCTX()
{
    return mp->ioctx;
}

std::string IRCBase::toMemberPrefix(const std::string& modes) const
{ // PREFIX=(qaohv)~&@%+
    const auto& prefixdef = mp->isupport.at("PREFIX"); // PREFIX is always present.

    const auto prefixmodes = prefixdef.substr(1, prefixdef.find_first_of(')') - 1);
    const auto prefix = prefixdef.substr(prefixdef.find_first_of(')') + 1);

    std::string ret;

    for (const char m : modes) {
        // Only accept a-zA-Z
        if (!(m > 'a' && m < 'z' || m > 'A' && m < 'Z')) { // De Morgan pls halp
            ret += ':';
            continue;
        }

        const auto ctpos = prefixmodes.find_first_of(m);
        if (ctpos == std::string::npos)
            ret += ':';
        else
            ret += prefix[ctpos];
    }

    return ret;
}

bool IRCBase::isChannelSymbol(char c)
{
    return mp->isChannelSymbol(c);
}

char IRCBase::channelModeGroup(char m) const
{
    return mp->channelModeGroup(m);
}

IRCError IRCBase::tryConnect()
{
    if (isOnline())
        return IRCError::AlreadyConnected;

    if (mp->hostname.empty())
        return IRCError::HostNotSet;

    if (mp->port.empty())
        return IRCError::PortNotSet;

    if (mp->ident.empty())
        return IRCError::IdentNotSet;

    if (mp->nickname.empty())
        return IRCError::NicknameNotSet;

    if (mp->realname.empty())
        return IRCError::RealnameNotSet;

    try {
        mp->sslExceptions.clear();
        mp->ioctx.restart();
        mp->endpoints = mp->resolver.resolve(mp->hostname, mp->port);
        if (mp->endpoints.empty())
            return IRCError::CannotResolveAddress;

        if (mp->useSSL) {
            mp->sslsock.emplace(mp->ioctx, mp->sslctx);

            mp->sslsock->set_verify_mode(asio::ssl::verify_peer);
            mp->sslsock->set_verify_callback([this](bool preverified, asio::ssl::verify_context& vc){
                if (preverified)
                    return true;

                X509* cert = X509_STORE_CTX_get_current_cert(vc.native_handle());
                mp->lastError = mp->verify_X509(cert);
                if (mp->lastError != IRCError::NoError) {
                    onConnectionError(mp->lastError);
                    return false;
                }
                else
                    return true;
            });

            asio::async_connect(mp->sslsock->lowest_layer(), mp->endpoints,
                [this](const asio::error_code& ec, const tcp::endpoint&) {
                    std::cout << "ssl connect error: " << ec.value() << " " << ec.message() << std::endl;
                    mp->lastError = SystemErrorToIRCError(ec);
                    if (ec) {
                        onConnectionError(mp->lastError);
                        return;
                    }
                    mp->sslsock->async_handshake(asio::ssl::stream_base::client,
                        [this](asio::error_code ec) {
                            mp->lastError = SystemErrorToIRCError(ec);
                            if (ec) {
                                onConnectionError(mp->lastError);
                                return;
                            }
                            // TODO  error codes!
                            std::cout << "ssl handshake error: " << ec.value() << " " << ec.message() << std::endl;
                            mp->connected(ec);
                        });
                });
        }
        else {
            mp->sock.emplace(mp->ioctx);
            asio::async_connect(mp->sock.value(), mp->endpoints,
                [this](const asio::error_code& ec, const tcp::endpoint&) {
                    mp->lastError = SystemErrorToIRCError(ec);
                    if (ec) {
                        onConnectionError(mp->lastError);
                        return;
                    }
                    mp->connected(ec);
                }
            );
        }
    }
    catch (const std::system_error& e) {
        return SystemErrorToIRCError(e);
    }

    return IRCError::NoError;
}

IRCError IRCBase::disconnectFromServer(const std::string& quitMessage)
{
    if (!mp->isConnected)
        return IRCError::NotConnected;

    if (mp->isOnline) {
        mp->write(Command::IRC::QUIT, quitMessage);
        // TODO  Quit timeout; if we never get disconnected by server, we forcefully must do so.
    }
    else {
        if (mp->useSSL)
            mp->sslsock->lowest_layer().close();
        else
            mp->sock->close();
    }

    if (mp->keepaliveFreq > std::chrono::seconds(0))
        mp->stopKeepaliveTimer();

    return IRCError::NoError;
}

IRCError IRCBase::lastErrorCode() const
{
    return mp->lastError;
}

bool IRCBase::isOnline() const
{
    return mp->isOnline;
}

bool IRCBase::isConnected() const
{
    return mp->isConnected;
}

bool IRCBase::isSSL() const
{
    return mp->useSSL;
}

const std::vector<std::string>& IRCBase::clientV3Support()
{
    return V3Support;
}

const std::vector<std::string>& IRCBase::enabledV3Support() const
{
    return mp->ircv3enabled;
}

const std::vector<std::string>& IRCBase::registeredV3Support() const
{
    return mp->registeredV3support;
}

const std::vector<std::string>& IRCBase::serverV3Support() const
{
    return mp->serverV3support;
}

const std::unordered_map<std::string, std::string>& IRCBase::isupport() const
{
    return mp->isupport;
}

const std::vector<std::shared_ptr<IRCChannel>>& IRCBase::channels() const
{
    return mp->channels;
}

std::shared_ptr<IRCChannel> IRCBase::getChannel(const std::string& name) const
{
    for (auto chanp : mp->channels)
        if (strEquals(chanp->name(), name))
            return chanp;
    return nullptr;
}

std::shared_ptr<IRCMember> IRCBase::getMember(const std::string& nickname) const
{
    for (auto memp : mp->allMembers)
        if (strEquals(memp->prefix().toString(), nickname))
            return memp;
    return nullptr;
}

std::pair<std::shared_ptr<DCC>, IRCError> IRCBase::initiateDCC()
{
    try {
        auto dcc{ std::make_shared<DCC>(*this, mp->ioctx) };
        return { dcc, IRCError::NoError };
    }
    catch (const std::exception& e) {
        std::cerr << "initiateDCC exception: " << e.what() << std::endl;
        return { nullptr, IRCError::DCC_InitiateError };
    }
}

const IRCMessage* IRCBase::getCurrentMessageRaw() const
{
    return mp->currentMessage;
}

void IRCBase::useIRCv3Features(const std::vector<std::string>& features)
{
    auto isFeatureSupported = [](const std::string& feature) {
        return std::find(V3Support.begin(), V3Support.end(), feature) != V3Support.end();
    };

    mp->ircv3enabled.clear();

    for (const auto& feature : features) {
        if (isFeatureSupported(feature))
            mp->ircv3enabled.emplace_back(feature);
    }
}
