/*
 * IdealIRC Core - Internet Relay Chat API
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "IRCMessage.h"
#include <algorithm>
#include <iterator>

IRCMessage::IRCMessage(const IRCPrefix& defaultSender, const std::string& message)
    : m_sender(defaultSender)
{
    /*
     * :sender command arg :message
    */

    if (message.empty())
        return;

    enum class S {
        IRCv3Tags,
        Prefix,
        Command,
        Argument,
        Message
    } parseState = S::IRCv3Tags;

    auto it = message.begin() + 1;

    if (message[0] != '@') {
        parseState = S::Prefix;
    }

    if (message[0] != ':' && message[0] != '@') {
        // Set up parser for the format: "cmd arg :msg"
        it = message.begin();
        parseState = S::Command;
    }

    /*
     * Tokenize the message. Three formats may occur:
     *  :server.addr cmd arg :msg
     *  @IRCv3Tag;tag2;tag3 :server.addr cmd arg :msg
     *  cmd arg :msg
    */
    while (it != message.end()) {
        auto next = std::find(it, message.end(), ' ');

        switch (parseState) {
            case S::IRCv3Tags:
            {
                std::string tagstr;
                std::copy(it, next, std::back_inserter(tagstr));
                parseIRCv3Tags(tagstr);
                parseState = S::Prefix;
                break;
            }

            case S::Prefix:
                /* This condition happens if parseState started on IRCv3Tags. */
                if (*it == ':')
                    ++it;

                m_sender = IRCPrefix(std::string(it, next));
                parseState = S::Command;
                break;

            case S::Command:
                m_command = std::string(it, next);
                parseState = S::Argument;
                break;

            case S::Argument:
                if (*it == ':') {
                    ++it;
                    parseState = S::Message;
                    // note: fallthrough to "Message"
                }
                else {
                    m_args.emplace_back(it, next);
                    break;
                }
                [[fallthrough]];

            case S::Message:
                next = message.end();
                m_message = std::string(it, next);
                break;
        }

        it = (next == message.end()) ? next : next + 1;
    }
}

const std::string& IRCMessage::operator[](int idx) const
{
    static const std::string empty;
    if (idx < 0 || idx >= m_args.size())
        return empty;
    else
        return m_args[idx];
}

void IRCMessage::parseIRCv3Tags(const std::string& tagstr)
{
    auto it = tagstr.begin();
    while (it != tagstr.end()) {
        auto next = std::find(it, tagstr.end(), ';');
        m_tags.emplace_back(it, next);
        it = (next == tagstr.end()) ? next : next + 1;
    }
}
