/*
 * IdealIRC Core - Internet Relay Chat API
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "IRCPrefix.h"
#include <fmt/format.h>
#include <algorithm>

IRCPrefix::IRCPrefix(const std::string& prefix)
{
    auto atIt = std::find(prefix.begin(), prefix.end(), '@');
    if (atIt != prefix.end()) {
        // nickname!user@host
        m_type = Type::user;
        auto exclmIt =  std::find(prefix.begin(), prefix.end(), '!');
        m_nickname = std::string(prefix.begin(), exclmIt);
        m_user = std::string(exclmIt + 1, atIt);
        m_host = std::string(atIt + 1, prefix.end());
    }
    else {
        m_type = Type::server;
        m_servername = prefix;
    }
}

IRCPrefix::IRCPrefix(const IRCPrefix& other)
    : m_servername(other.m_servername)
    , m_nickname(other.m_nickname)
    , m_user(other.m_user)
    , m_host(other.m_host)
    , m_type(other.m_type)
{}

IRCPrefix::IRCPrefix(IRCPrefix&& other) noexcept
    : m_servername(std::move(other.m_servername))
    , m_nickname(std::move(other.m_nickname))
    , m_user(std::move(other.m_user))
    , m_host(std::move(other.m_host))
    , m_type(std::move(other.m_type))
{}

IRCPrefix::IRCPrefix(IRCPrefix::Type t)
    : m_type(t)
{}

IRCPrefix& IRCPrefix::operator=(const IRCPrefix& other)
{
    m_servername = other.m_servername;
    m_nickname = other.m_nickname;
    m_user = other.m_user;
    m_host = other.m_host;
    m_type = other.m_type;
    return *this;
}

const std::string& IRCPrefix::toString() const
{
    if (m_type == Type::server)
        return m_servername;
    else
        return m_nickname;
}

const std::string& IRCPrefix::servername() const
{
    return m_servername;
}

const std::string& IRCPrefix::nickname() const
{
    return m_nickname;
}

const std::string& IRCPrefix::user() const
{
    return m_user;
}

const std::string& IRCPrefix::host() const
{
    return m_host;
}

void IRCPrefix::setNickname(const std::string& nickname)
{
    m_nickname = nickname;
}

void IRCPrefix::setHost(const std::string& host)
{
    m_host = host;
}

std::string IRCPrefix::composite() const
{
    if (m_type == Type::server)
        return m_servername;
    else
        return fmt::format("{}!{}@{}", m_nickname, m_user, m_host);
}

IRCPrefix::Type IRCPrefix::type() const
{
    return m_type;
}

IRCPrefix IRCPrefix::fromNickname(const std::string& nickname)
{
    IRCPrefix p(Type::user);
    p.m_nickname = nickname;
    return p;
}

bool IRCPrefix::operator!=(const IRCPrefix& other)
{
    return other.type() != m_type || other.toString() != toString();
}

bool IRCPrefix::operator==(const IRCPrefix& other)
{
    return other.type() == m_type && other.toString() == toString();
}

bool operator!=(const IRCPrefix& lhs, const IRCPrefix& rhs)
{
    return lhs.type() != rhs.type() || lhs.toString() != rhs.toString();
}

bool operator==(const IRCPrefix& lhs, const IRCPrefix& rhs)
{
    return lhs.type() == rhs.type() && lhs.toString() == rhs.toString();
}
