/*
 * IdealIRC Core - Internet Relay Chat API
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "../IRCBasePriv.h"
#include "../Utilities.h"

#include <asio/ssl.hpp>

IRCError IRCBasePriv::verify_X509(X509* cert)
{
    char sn_char[256];
    X509_NAME_oneline(X509_get_subject_name(cert), sn_char, 256);
    std::string subject(sn_char);

    std::string CN;
    {
        auto pos = subject.find("/CN=");
        if (pos != std::string::npos) {
            auto begin = subject.begin() + pos;
            auto end = std::find(begin + 1, subject.end(), '/');
            CN = std::string(begin, end);
        }
    }

    if (X509_check_issued(cert, cert) == X509_V_OK) {
        if (sslExcept.selfSigned)
            sslExceptions.emplace_back(IRCError::SSL_SelfSigned);
        else
            return IRCError::SSL_SelfSigned;
    }

    const auto notbefore = X509_getm_notBefore(cert);
    if (X509_cmp_current_time(notbefore) > -1) {
        if (sslExcept.expired)
            sslExceptions.emplace_back(IRCError::SSL_NotYetValid);
        else
            return IRCError::SSL_NotYetValid;
    }

    const auto notafter = X509_getm_notAfter(cert);
    if (X509_cmp_current_time(notafter) < 1) {
        if (sslExcept.expired)
            sslExceptions.emplace_back(IRCError::SSL_Expired);
        else
            return IRCError::SSL_Expired;
    }

    if (CN.empty()) {
        if (sslExcept.CNMismactch)
            sslExceptions.emplace_back(IRCError::SSL_CN_Mismatch);
        else
            return IRCError::SSL_CN_Missing;
    }

    /*
     * CN wildcard matching.
     * Rules picked from: https://en.wikipedia.org/wiki/Wildcard_certificate#Examples
     * TODO There are probably more rules! Remember this is IRC and not HTTP so the rules may be more relaxed.
    */
    if (std::find(CN.begin(), CN.end(), '*') != CN.end()) {

        // Lambda: Count occurrences of character in a string
        auto strCount = [](const std::string& str, char c) -> std::size_t {
            auto count = 0;
            std::for_each(str.begin(), str.end(), [&count,c](char ch) { if (c == ch) ++count; });
            return count;
        };

        // Reject "*" and wildcard on the end of CN
        if (CN.back() == '*')
            return IRCError::SSL_CN_WildcardIllegal;

        // Reject multiple wildcards
        if (strCount(CN, '*') > 1)
            return IRCError::SSL_CN_WildcardIllegal;

        /*
         * Reject "*.tld"
         * Actually, reject if there is only a single dot
        */
        if (strCount(CN, '.') == 1)
            return IRCError::SSL_CN_WildcardIllegal;

        // Reject sub.*.domain.com
        auto wcit = std::find(CN.begin(), CN.end(), '*');
        if (wcit != CN.begin()) {
            if (*(wcit-1) == '.' && *(wcit+1) == '.')
                return IRCError::SSL_CN_WildcardIllegal;
        }

        if (!singleWildcardMatch(hostname, CN)) {
            if (sslExcept.CNMismactch)
                sslExceptions.emplace_back(IRCError::SSL_CN_Mismatch);
            else
                return IRCError::SSL_CN_Mismatch;
        }
    }

        /* CN exact hostname match */
    else if (CN != hostname) {
        if (sslExcept.CNMismactch)
            sslExceptions.emplace_back(IRCError::SSL_CN_Mismatch);
        else
            return IRCError::SSL_CN_Mismatch;
    }

    return IRCError::NoError;
}