/*
 * IdealIRC Core - Internet Relay Chat API
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "Utilities.h"

#if defined(WIN32) || defined(WIN64)
#include <winsock2.h>
    #include <iphlpapi.h>
    #include <cstdlib>
#endif

#include <fmt/format.h>
#include <cstdint>
#include <cstdio>
#include <sstream>
#include <vector>

namespace {
constexpr auto DefaultIpAddress{ "127.0.0.1" };

std::string getIpFromSecondToken(const std::string& input)
{
    if (input.empty())
        return {};

    /*
     * Simple shell injection check, input is likely to come from a popen() (shell) call.
     * There is no way to tell if the called program has been tampered with...
    */
    auto it = std::find_if(begin(input), end(input),
                           [](char c) {
                               return c == '|' || c == '&' || c == ';' || c == '$';
                           });

    if (it != input.end())
        return DefaultIpAddress; // someone's fooling around with 'ip' command, let's not continue...
    else {
        auto start = input.find(' ') + 1;
        auto len = input.find(' ', start);
        if (len != std::string::npos)
            len -= start;
        return input.substr(start, len);
    }
}

#if defined(__linux__) || defined(__linux)
std::string getIpAddress_Linux()
{
    std::string ipaddr{ DefaultIpAddress };
    std::string iface;

    {
        char buf[64]{};
        // This command will return either empty or something like "dev eth0 some more data here" - we want the eth0 part.
        auto fd = popen(R"(ip route show default | grep -E -o 'dev\s.+(\s|$)')", "r");
        auto readres = fgets(buf, sizeof(buf), fd);
        pclose(fd);

        if (!readres)
            return ipaddr;

        iface.assign(buf, sizeof(buf));
        iface = getIpFromSecondToken(iface);
    }

    if (!iface.empty()) {
        char buf[64]{};
        // This command will match either empty or exactly "inet 123.45.67.89" though with a proper IP address.
        static constexpr auto regex{ R"(inet\s[[:digit:]]{1,3}\.[[:digit:]]{1,3}\.[[:digit:]]{1,3}\.[[:digit:]]{1,3})" };
        const auto cmd{ fmt::format("ip address show {} | grep -E  -o '{}'", iface, regex) };
        auto fd = popen(cmd.c_str(), "r");
        auto readres = fgets(buf, sizeof(buf) - 1, fd);
        pclose(fd);

        if (!readres)
            return ipaddr;

        // We use strlen() since buf will hold many zero-bytes and popen() doesn't give us a size.
        ipaddr.assign(buf, strlen(buf));

        auto start = ipaddr.find(' ') + 1;
        ipaddr = ipaddr.substr(start);

        if (ipaddr.back() == '\n')
            ipaddr.erase(ipaddr.end() - 1);
    }

    return ipaddr;
}
#endif

#if defined(WIN32) || defined(WIN64)
std::string getIpAddress_Windows()
{
    // This function is based on the example given here:
    // https://docs.microsoft.com/en-us/windows/win32/api/iphlpapi/nf-iphlpapi-getadaptersinfo

    std::string ipaddr{ DefaultIpAddress };

    /* Get the index of the adapter with default route (0.0.0.0) */
    DWORD adpidx{};
    if (GetBestInterface(0, &adpidx) != NO_ERROR)
        return ipaddr;

    PIP_ADAPTER_INFO pAdapterInfo;
    PIP_ADAPTER_INFO pAdapter = NULL;

    ULONG ulOutBufLen = sizeof(IP_ADAPTER_INFO);
    pAdapterInfo = (IP_ADAPTER_INFO*)malloc(sizeof(IP_ADAPTER_INFO));
    if (!pAdapterInfo)
        return ipaddr;

    /* Make an initial call to GetAdaptersInfo to get the necessary size into the ulOutBufLen variable */
    if (GetAdaptersInfo(pAdapterInfo, &ulOutBufLen) == ERROR_BUFFER_OVERFLOW) {
        free(pAdapterInfo);
        pAdapterInfo = (IP_ADAPTER_INFO*)malloc(ulOutBufLen);
        if (!pAdapterInfo)
            return ipaddr;
    }

    /* Actual call for adapter info */
    if (GetAdaptersInfo(pAdapterInfo, &ulOutBufLen) != NO_ERROR)
        return ipaddr;

    for (pAdapter = pAdapterInfo; pAdapter; pAdapter = pAdapter->Next) {
        if (pAdapter->Index == adpidx) {
            ipaddr = pAdapter->IpAddressList.IpAddress.String;
            break;
        }
    }

    if (pAdapterInfo)
        free(pAdapterInfo);

    return ipaddr;
}
#endif

} // anonymous namespace

std::pair<std::string,std::string> FormatCTCPLine(std::string line)
{
    /* Clean away the CTCP byte flags */
    if (line[0] == CTCPflag)
        line.erase(line.begin());

    auto endIt = std::find(line.begin(), line.end(), CTCPflag);
    if (endIt != line.end())
        line.pop_back();

    auto cmdIt = std::find(line.begin(), line.end(), ' ');
    std::string cmd(line.begin(), cmdIt);
    std::string ctcpMsg;
    if (cmdIt != line.end())
        ctcpMsg = std::string(cmdIt + 1, line.end());

    return std::make_pair(cmd, ctcpMsg);
}

/*!
 * @brief Simple wildcard match, supporting only a single wildcard.
 * @returns true if string 'str' matches with the wildcard string 'wc'.
*/
bool singleWildcardMatch(const std::string& str, const std::string& wc)
{
    if (str.empty() || wc.empty())
        return false;

    auto wcit = std::find(wc.begin(), wc.end(), '*');
    if (wcit == wc.end())
        return str == wc;

    /*
     * Since we only support a single wildcard we only need to match the left side and right side
     * of the asterisk (wildcard) against the target string 'str'.
     * Also if the "end" of left match is equal to the "begin" of right match, it's a mismatch.
     */

    bool hasLeft, hasRight { false };
    auto leftEnd = str.end();
    auto rightBegin = str.end();

    /* Left side */
    if (wcit != wc.begin()) {
        std::string wildls(wc.begin(), wcit);
        auto strmatch = str.substr(0, wildls.length());
        if (wildls != strmatch)
            return false;
        leftEnd = str.begin() + wildls.length();
        hasLeft = true;
    }

    /* Right side */
    if (wcit != wc.end() - 1) {
        std::string wildrs(wcit + 1, wc.end());
        auto mpos = str.find(wildrs);
        if (mpos == std::string::npos)
            return false;
        auto strmatch = str.substr(mpos);
        if (wildrs != strmatch)
            return false;
        rightBegin = str.begin() + mpos;
        hasRight = true;
    }

    // Avoid matching "left*right" with "leftright"
    return !(hasLeft && hasRight && leftEnd == rightBegin);
}

std::string longIpToNormal(std::uint32_t longip)
{
    unsigned ip = longip;
    unsigned a = (ip & 0xFF000000u) >> 24u;
    unsigned b = (ip & 0x00FF0000u) >> 16u;
    unsigned c = (ip & 0x0000FF00u) >> 8u;
    unsigned d = ip & 0x000000FFu;
    return fmt::format("{}.{}.{}.{}", a, b, c, d);
}

std::string longIpToNormal(const std::string& longip)
{
    try {
        return longIpToNormal( std::stoul(longip) );
    }
    catch (...) {
        return "0.0.0.0";
    }
}

std::uint32_t normalIpToLong(const std::string& normalip)
{
    std::stringstream ss(normalip);
    std::string p;
    std::vector<unsigned> parts;
    while (getline(ss, p, '.'))
        parts.emplace_back(std::stoul(p));

    uint32_t ret = parts[0];
    for (int i = 1; i < 4; ++i) {
        ret <<= 8u;
        ret |= parts[i];
    }

    return ret;
}

std::string concatenateModes(const std::unordered_map<char, std::string>& modes)
{
    std::string ret;
    std::string modeargs;

    for (const auto& [m,a] : modes) {
        ret += m;
        if (!a.empty()) {
            if (!modeargs.empty())
                modeargs += ' ';
            modeargs += a;
        }
    }

    if (!modeargs.empty())
        ret += " " + modeargs;

    return ret;
}

bool strEquals(const std::string& l, const std::string& r)
{
    if (l.size() != r.size())
        return false;

    return std::equal(l.begin(), l.end(), r.begin(),
                      [](char cl, char cr) {
                          return toupper(cl) == toupper(cr);
                      }
    );
}

std::string toBase64(const std::string& input)
{
    static const char b64map[] = { 'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P',
                                   'Q','R','S','T','U','V','W','X','Y','Z','a','b','c','d','e','f',
                                   'g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v',
                                   'w','x','y','z','0','1','2','3','4','5','6','7','8','9','+','/' };

    std::string ret;

    for (auto i = 0; i < input.size(); i += 3) {
        std::int_fast32_t seq{}; // Need only 24 bits of this
        std::int_fast32_t a{}, b{}, c{}, d{};
        bool cBlank{}, dBlank{};

        seq |= input[i];
        seq <<= 8;

        if (i + 1 < input.size())
            seq |= input[i + 1];
        else
            cBlank = true;
        seq <<= 8;

        if (i + 2 < input.size())
            seq |= input[i + 2];
        else
            dBlank = true;

        a = (seq & 0xFC0000) >> 18;
        b = (seq & 0x3F000) >> 12;
        c = (seq & 0xFC0) >> 6;
        d = seq & 0x3F;

        ret.push_back(b64map[a]);
        ret.push_back(b64map[b]);

        if (cBlank)
            ret.push_back('=');
        else
            ret.push_back(b64map[c]);

        if (dBlank)
            ret.push_back('=');
        else
            ret.push_back(b64map[d]);
    }

    return ret;
}

std::string findOwnIpAddress()
{
#if defined(__linux__) || defined(__linux)
    auto ipaddr = getIpAddress_Linux();
#elif defined(WIN32) || defined(WIN64)
    auto ipaddr = getIpAddress_Windows();
#else
    #error Unsupported platform, implementation of findOwnIpAddress() missing!
#endif
    return ipaddr.empty() ? DefaultIpAddress : ipaddr;
}

bool isPrivateIp(const std::string& ip)
{
    /*
     * Class A: 10.0.0.0 - 10.255.255.255
     * Class B: 172.16.0.0 - 172.31.255.255
     * Class C: 192.168.0.0 - 192.168.255.255
    */

    const auto intip{ normalIpToLong(ip) };

    const auto classIp{ (intip & 0xFF000000) >> 24 };

    /* Class A */
    if (classIp == 10) {
        // All addresses starting on "10." is in fact private.
        return true;
    }

    /* Class B */
    else if (classIp == 172) {
        // Check if second byte is between 16 and 31.
        const auto num = (intip & 0x00FF0000) >> 16;
        return (num >= 16 && num <= 31);
    }

    /* Class C */
    else if (classIp == 192) {
        // Check if second byte is exactly 168.
        return ((intip & 0x00FF0000) >> 16) == 168;
    }

    /* Not in any of the classes, not a private IP. */
    else {
        return false;
    }
}
