/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2022  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "IWinDCCChat.h"
#include "IWinStatus.h"
#include <IRCClient/IRCBase.h>
#include <QMessageBox>
#include <iostream>

IWinDCCChat::IWinDCCChat(IWinStatus* statusParent, std::shared_ptr<DCC> dcc_, const QString& targetNickname)
    : IWin(IWin::Type::DCCChat, statusParent)
    , dcc(std::move(dcc_))
{
    setButtonText(targetNickname);

    layout = new QVBoxLayout();
    view = new IIRCView();
    input = new ILineEdit([this](int a, QString& b){ return tabComplete(a,b); });

    layout->setMargin(0);
    layout->setSpacing(2);
    layout->addWidget(view);
    layout->addWidget(input);
    setLayout(layout);
    setTabOrder(input, view);

    view->setFocusPolicy(Qt::FocusPolicy::NoFocus);

    connect(input, &ILineEdit::newLine,
            this, &IWinDCCChat::newLine);

    setupCallbacks();
}

bool IWinDCCChat::printWithCustomTime(const QDateTime& timestamp, PrintType ptype, const QString& text)
{
    view->print(timestamp, ptype, text);
    return true;
}

void IWinDCCChat::refreshWindowTitle()
{
    setWindowTitle(tr("DCC Chat with: %1").arg(getButtonText()));
}

void IWinDCCChat::acceptAsTarget()
{
    const auto ip = QString::fromStdString(dcc->ip());

    if (dcc->isReversed()) {
        print(PrintType::ProgramInfo, tr("Waiting for participant..."));
    }
    else {
        const auto port = QString::number(dcc->port());
        print(PrintType::ProgramInfo, tr("Connecting to %1:%2...").arg(ip, port));
    }

    dcc->accept(getButtonText().toStdString());
}

void IWinDCCChat::setupCallbacks()
{
    dcc->setCallbackConnected([this]{ onConnected(); });
    dcc->setCallbackDisconnected([this](IRCError e){ onDisconnected(e); });
    dcc->setCallbackRead([this](const DCC::ByteString& data, std::size_t length){ onRead(data, length); });
}

void IWinDCCChat::closeEvent(QCloseEvent* evt)
{
    auto btn = QMessageBox::question(this, tr("Disconnect DCC Chat"),
                                           tr("Closing this window will also disconnect the DCC session, and no more messages can be received over this conversation.\nProceed?"));

    if (btn == QMessageBox::Yes) {
        dcc->resetCallbacks();
        dcc->disconnect();

        evt->accept();

        IWin::closeEvent(evt);
    }
    else {
        evt->ignore();
    }
}

void IWinDCCChat::onConnected()
{
    print(PrintType::ProgramInfo, tr("Connected to remote peer"));
}

void IWinDCCChat::onDisconnected(IRCError e)
{
    print(PrintType::ProgramInfo, tr("Disconnected from remote peer"));
}

void IWinDCCChat::newLine(const QString& line)
{
    if (dcc->isConnected()) {
        const auto nickname{ QString::fromStdString(dcc->context().getNickname()) };
        print(PrintType::OwnText, tr("<%1> %2").arg(nickname, line));

        auto data = DCC::stringToByteString(line.toStdString());
        data.push_back(std::byte('\r'));
        data.push_back(std::byte('\n'));
        dcc->write(data);
    }
    else {
        print(PrintType::ProgramInfo, tr("Not connected with remote peer"));
    }
}

void IWinDCCChat::onRead(const DCC::ByteString& data, std::size_t length)
{
    auto beginIt = data.cbegin();
    auto end = data.cbegin() + static_cast<int>(length);

    while (true) {
        auto endIt = std::find_if(beginIt, end,
                                  [](std::byte byte) { return byte == std::byte('\n'); });

        std::basic_string<std::byte> basictemp(beginIt, endIt);
        std::string temp(reinterpret_cast<char*>(basictemp.data()), basictemp.size());

        if (endIt == end) {
            lineBuffer += QString::fromStdString(temp);
            break;
        }

        lineBuffer += QString::fromStdString(temp);
        if (!lineBuffer.isEmpty() && lineBuffer.back() == '\r')
            lineBuffer.remove(lineBuffer.length()-1, 1);

        print(PrintType::Normal, tr("<%1> %2").arg(getButtonText(), lineBuffer));

        lineBuffer.clear();

        beginIt = ++endIt;
    }
}

int IWinDCCChat::tabComplete(int index, QString& pattern)
{
    return 0;
}

void IWinDCCChat::disconnectFromDCC()
{
    dcc->disconnect();
}

void IWinDCCChat::initiateReversed(std::shared_ptr<DCC> dcc_)
{
    reverseInitiatePending = false;
    dcc = std::move(dcc_);
    setupCallbacks();
    dcc->accept(getButtonText().toStdString());
}
