/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "IWinStatus.h"
#include "IWinChannel.h"
#include "IWinDCCChat.h"
#include "MdiManager.h"
#include "Script/Manager.h"
#include "IRC.h"
#include "ConfigMgr.h"
#include <QMessageBox>

int IWinStatus::StatusWindowCount = 0;

IWinStatus::IWinStatus(const QString& buttonText)
    : IWin(Type::Status)
    , connection(*this)
{
    setButtonText(buttonText);
    ++StatusWindowCount;

    layout = new QVBoxLayout();
    view = new IIRCView();
    input = new ILineEdit([this](int a, QString& b){ return tabComplete(a,b); });

    static int connectionIdCount{ 0 };
    connect(view, &IIRCView::customContextMenuRequested, [this](const QPoint& point){
        menuSymbols.set("cid", ++connectionIdCount);
        ScriptManager::instance()->contextMenuPopup(ScriptMenuType::Status, mapToGlobal(point), menuSymbols);
    });

    layout->setMargin(0);
    layout->setSpacing(2);
    layout->addWidget(view);
    layout->addWidget(input);
    setLayout(layout);
    setTabOrder(input, view);

    view->setFocusPolicy(Qt::FocusPolicy::NoFocus);

    connect(input, &ILineEdit::newLine,
            this, &IWinStatus::newLine);

    connect(&connection, &IRC::connected, [this] {
        MdiManager::instance().connectionStateChange();
        refreshWindowTitle();
    });

    connect(&connection, &IRC::registered, [this] {
        refreshWindowTitle();
    });

    connect(&connection, &IRC::disconnected, [this] {
        disconnectedFromServer();
        MdiManager::instance().connectionStateChange();
        refreshWindowTitle();
    });

    Qt_Eventloop_Hook();
}

bool IWinStatus::printWithCustomTime(const QDateTime& timestamp, PrintType ptype, const QString& text)
{
    view->print(timestamp, ptype, text);
    MdiManager::instance().highlight(this, HL_Activity);
    
    return true;
}

void IWinStatus::refreshWindowTitle()
{
    const auto& isupport = connection.isupport();
    const auto it = isupport.find("NETWORK");
    QString title;

    if (!connection.isOnline() && !connection.isConnected())
        title = tr("Status (Offline)");
    else if (!connection.isOnline() && connection.isConnected())
        title = tr("Status (Registering)");
    else if (it == isupport.end())
        title = QStringLiteral("%1 (%2)")
                        .arg(QString::fromStdString( connection.getNickname() ))
                        .arg(QString::fromStdString( connection.getHostname() ));
    else
        title = QStringLiteral("%1 (%2)")
                        .arg(QString::fromStdString( connection.getNickname() ))
                        .arg(QString::fromStdString( it->second ));

    setWindowTitle(title);
    MdiManager::instance().renameWindowButton(this, title);
}

void IWinStatus::printTo(const QString& target, const PrintType ptype, const QString& text)
{
    MdiManager::instance().print(this, target, ptype, text);
}

void IWinStatus::printToTypes(const IWin::Type toType, const PrintType ptype, const QString& text)
{
    MdiManager::instance().printToTypes(this, toType, ptype, text);
}

void IWinStatus::printToActive(const PrintType ptype, const QString& text)
{
    MdiManager::instance().printToActive(this, ptype, text);
}

void IWinStatus::printToAll(const PrintType ptype, const QString& text)
{
    MdiManager::instance().printToAll(this, ptype, text);
}

IWinChannel* IWinStatus::createChannelWindow(const QString& name)
{
    auto& mdiManager = MdiManager::instance();
    IWin* subwinBase = mdiManager.findWindow(this, name);
    if (!subwinBase)
        subwinBase = mdiManager.createSubwindow(this, name, IWin::Type::Channel);

    return dynamic_cast<IWinChannel*>(subwinBase);
}

IWin* IWinStatus::createPrivateWindow(const QString& target, bool setFocus, Highlight buttonHighlight)
{
    auto& mdiManager = MdiManager::instance();
    IWin* subwinBase = mdiManager.findWindow(this, target);
    if (!subwinBase)
        subwinBase = mdiManager.createSubwindow(this, target, IWin::Type::Private, setFocus, buttonHighlight);

    return subwinBase;
}

IWinDCCChat* IWinStatus::createDCCChatWindow(std::shared_ptr<DCC> dcc, const IRCPrefix& sender)
{
    auto& mdiManager = MdiManager::instance();
    auto* win = mdiManager.createDCCSubwindow(this, IWin::Type::DCCChat, std::move(dcc), sender);
    return dynamic_cast<IWinDCCChat*>(win);
}

IWin* IWinStatus::getActiveWindow()
{
    auto& mdiManager = MdiManager::instance();
    if (mdiManager.currentWindow()->getStatusParent() != this)
        return this;
    return mdiManager.currentWindow();
}

QList<IWin*> IWinStatus::subWindows() const
{
    return MdiManager::instance().childrenOf(this);
}

void IWinStatus::closeEvent(QCloseEvent* evt)
{
    auto& mdiManager = MdiManager::instance();
    const auto& conf = ConfigMgr::instance();

    if (StatusWindowCount == 1 ||
        (connection.isConnected() && QMessageBox::question(this, tr("Close status window"), tr("This status window has a connection. Close?"))
                                   == QMessageBox::No)) {
        evt->ignore();
        return;
    }

    if (connection.isConnected()) {
        connect(&connection, &IRC::disconnected, [&] { closeAllChildren(); mdiManager.toMdiwin(this)->close(); });
        const auto quitMsg = conf.common("QuitMessage").toStdString();
        const auto e = connection.disconnectFromServer(quitMsg);
        if (e != IRCError::NoError)
            qWarning() << "disconnectFromServer returned:" << QString::fromStdString(IRCErrorToString(e));
        evt->ignore();
        return;
    }

    --StatusWindowCount;

    IWin::closeEvent(evt);
}

void IWinStatus::newLine(const QString& line)
{
    InputHandler inHndl(connection);
    inHndl.parse(*this, line);
}

void IWinStatus::closeAllChildren()
{
    auto& mdiManager = MdiManager::instance();
    QList<QMdiSubWindow*> subwinList = mdiManager.mdiChildrenOf(this);
    for (QMdiSubWindow* subwin : subwinList)
        subwin->close();
}

void IWinStatus::Qt_Eventloop_Hook()
{
    const int nextInterval = connection.poll() ? 0 : 10;
    QTimer::singleShot(nextInterval, this, std::bind(&IWinStatus::Qt_Eventloop_Hook, this));
}

void IWinStatus::disconnectedFromServer()
{
    const auto printDisconnected = [this](IWin* sw){
        sw->print(PrintType::ProgramInfo, tr("Disconnected"));
    };

    printDisconnected(this);

    auto subwins = subWindows();
    for (auto* sw : subwins) {
        if (sw->getType() == IWin::Type::Channel) {
            auto* chan = dynamic_cast<IWinChannel*>(sw);
            chan->resetNicklist();
            printDisconnected(chan);
        }
        else if (sw->getType() == IWin::Type::Private) {
            printDisconnected(sw);
        }
    }
}

int IWinStatus::tabComplete(int index, QString& pattern)
{
    if (pattern.isEmpty() || !connection.isOnline())
        return 0;

    if (connection.isChannelSymbol( pattern[0].toLatin1() ))
        return connection.channelAutoComplete(index, pattern);

    return 0;
}
