/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "ButtonbarMgr.h"
#include "MdiManager.h"
#include "ConfigMgr.h"

int ButtonbarMgr::WindowTypeToGroupPos(const IWin* subwin)
{
    switch (subwin->getType()) {
    case IWin::Type::Status:
        return 1;
    case IWin::Type::Channel:
        return 2;
    case IWin::Type::Private:
        return 3;
    default:
        return 0;
    }
}

ButtonbarMgr::ButtonbarMgr(QToolBar *parent)
    : QObject(parent)
    , m_buttonBar(*parent)
{
    m_buttonBar.setContextMenuPolicy(Qt::CustomContextMenu);
    m_buttonBar.setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    m_buttonBar.setIconSize(QSize(16, 16));
    m_buttonBar.setFloatable(false);

    connect(&m_buttonBar, &QToolBar::customContextMenuRequested, this, &ButtonbarMgr::buttonBarContextMenu);
    m_others_rightSep = m_buttonBar.addSeparator();

    connect(&m_buttonBar, &QToolBar::orientationChanged, [this](Qt::Orientation o){
        QHashIterator<IWin*,QAction*> it(m_winbtn);
        while (it.hasNext()) {
            auto* action = it.next().value();
            auto* widget = m_buttonBar.widgetForAction(action);
            if (o == Qt::Vertical)
                widget->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
            else
                widget->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);
        }
    });
}

void ButtonbarMgr::addButton(IWin* subwin, QMdiSubWindow* mdiwin)
{
    auto* nn = findNextNeighbour(subwin);
    auto btn = Button(mdiwin);

    btn.toolButton->setMaximumWidth(150);
    if (m_buttonBar.orientation() == Qt::Vertical)
        btn.toolButton->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Fixed);
    else
        btn.toolButton->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);

    btn.action = m_buttonBar.insertWidget(nn, btn.toolButton);
    m_winbtn.insert(subwin, btn.action);

    buttons(subwin).push_back(btn);
    auto btnWidget = buttons(subwin).back().toolButton;

    connect(btnWidget, &QToolButton::clicked, [this,btnWidget,mdiwin](bool checked){
        if (!checked) {
            btnWidget->setChecked(true);
            return;
        }
        emit changeWindow(mdiwin);
    });
}

void ButtonbarMgr::delButton(IWin* subwin)
{
    auto clear = [this](Button& button){
        m_buttonBar.removeAction(button.action);
        button.menu->deleteLater();
        button.action->deleteLater();
        button.toolButton->deleteLater();
    };

    m_winbtn.remove(subwin);
    if (subwin->getType() != IWin::Type::Status
        && subwin->getType() != IWin::Type::Channel
        && subwin->getType() != IWin::Type::Private)
    {
        int otherPos = buttonPosition(m_others, subwin);
        if (otherPos > -1) {
            clear(m_others[otherPos]);
            m_others.removeAt(otherPos);
        }
        return;
    }

    if (subwin->getType() == IWin::Type::Status) {
        deleteGroup(subwin);
        return;
    }

    ButtonGroup* group = tryFindGroup(subwin);
    if (!group) {
        return;
    }

    for (int i = 0; i < group->size; ++i) {
        QList<Button>& buttons = (*group)[i];
        int buttonPos = buttonPosition(buttons, subwin);
        if (buttonPos < 0) continue;

        Button& button = buttons[buttonPos];
        if (button.subwin->widget() == subwin) {
            clear(button);
            buttons.removeAt(buttonPos);
            break;
        }
    }
}

void ButtonbarMgr::reloadButtonName(IWin* subwin, const QString& buttonText)
{
    Button* button = findButton(subwin);
    if (button) {
        button->toolButton->setText(buttonText);
        button->toolButton->setToolTip(buttonText);
        button->menuHead->setText(buttonText);
    }
}

void ButtonbarMgr::subwinActivated(IWin* prev, IWin* next)
{
    if (prev) {
        Button* button{ findButton(prev) };
        if (button)
            button->toolButton->setChecked(false);
    }
    if (next) {
        Button* button{ findButton(next) };
        if (button) {
            button->toolButton->setChecked(true);
            highlight(next, HL_None);
        }
    }
}

void ButtonbarMgr::highlight(IWin* window, Highlight state)
{
    auto& mdi = MdiManager::instance();
    if (window == mdi.currentWindow())
        state = HL_None;

    auto& conf = ConfigMgr::instance();
    QColor col { "#000000" };
    switch (state) {
        case HL_None:
            col = conf.color("WindowButtonNormal");
            break;
        case HL_Activity:
            col = conf.color("WindowButtonActivity");
            break;
        case HL_Message:
            col = conf.color("WindowButtonMessage");
            break;
        case HL_Attention:
            col = conf.color("WindowButtonAttention");
            break;
    }

    auto* btn = findButton(window);
    if (btn && (state > btn->hlState || state == HL_None)) {
        btn->hlState = state;
        auto palette = btn->toolButton->palette();
        palette.setColor(QPalette::ButtonText, QColor(col));
        btn->toolButton->setPalette(palette);
    }
}

void ButtonbarMgr::buttonBarContextMenu(const QPoint& pos)
{
    QAction* action = m_buttonBar.actionAt(pos);
    if (!action)
        return;

    IWin *subwin = m_winbtn.key(action);
    if (!subwin)
        return;

    Button* button = findButton(subwin);
    if (!button)
        return;

    QPoint gpos = m_buttonBar.mapToGlobal(pos);
    button->menu->popup(gpos);
}

ButtonbarMgr::ButtonGroup* ButtonbarMgr::tryFindGroup(IWin* subwin)
{
    auto lcontains = [](const QList<Button>& list, IWin* sw){
        for (const auto& item : list) {
            if (item.subwin->widget() == sw)
                return true;
        }
        return false;
    };

    ButtonGroup* null_grp{ nullptr };
    for (auto& group : m_groups) {
        for (int i = 0; i < group.size; ++i) {
            auto& list = group[i];
            if (lcontains(list, subwin))
                return &group;
        }
    }
    return null_grp;
}

QList<ButtonbarMgr::Button>& ButtonbarMgr::buttons(IWin* subwin)
{
    if (subwin->getType() != IWin::Type::Status
        && subwin->getType() != IWin::Type::Channel
        && subwin->getType() != IWin::Type::Private)
    {
        return m_others;
    }

    ButtonGroup* group = tryFindGroup(subwin);
    if (group)
        return (*group)[static_cast<int>(subwin->getType())-1];

    if (subwin->getType() == IWin::Type::Status) {
        m_groups.emplace_back();
        m_groups.back().rightSep = m_buttonBar.addSeparator();
        return m_groups.back().status;
    }

    IWin* parent = subwin->getParent();
    group = tryFindGroup(parent);
    if (!group)
        return m_others;

    if (subwin->getType() == IWin::Type::Channel)
        return group->channel;

    else // Implicit "Private"
        return group->privmsg;
}

QAction* ButtonbarMgr::findNextNeighbour(IWin* subwin)
{
    if (subwin->getType() == IWin::Type::Status) {
        auto* group = tryFindGroup(subwin);
        return group ? group->rightSep : nullptr;
    }
    else if (subwin->getType() == IWin::Type::Channel) {
        auto* group = tryFindGroup(subwin->getParent());
        if (group->privmsg.isEmpty())
            return group->rightSep;
        else
            return group->privmsg.front().action;
    }
    else if (subwin->getType() == IWin::Type::Private) {
        return tryFindGroup(subwin->getParent())->rightSep;
    }
    else {
        return m_others_rightSep;
    }
}

ButtonbarMgr::Button* ButtonbarMgr::findButton(IWin* subwin)
{
    if (!subwin)
        return nullptr;
    else {
        ButtonGroup* group = tryFindGroup(subwin);
        if (group) {
            for (int i = 0; i < group->size; ++i) {
                QList<Button>& list = (*group)[i];
                int bpos = buttonPosition(list, subwin);
                if (bpos > -1)
                    return &list[bpos];
            }
        }
    }

    int otherPos = buttonPosition(m_others, subwin);
    if (otherPos < 0)
        return nullptr;
    else {
        Button& button = m_others[otherPos];
        return &button;
    }
}

void ButtonbarMgr::deleteGroup(IWin* statuswin)
{
    auto it = m_groups.begin();
    for (; it != m_groups.end(); ++it) {
        if ((*it).status[0].subwin->widget() == qobject_cast<QWidget*>(statuswin))
            break;
    }
    ButtonGroup& group = *it;
    for (int i = 0; i < group.size; ++i) {
        QList<Button>& list = group[i];
        for (auto& item : list) {
            item.menuHead->deleteLater();
            item.menuSep->deleteLater();
            item.menuClose->deleteLater();
            item.menu->deleteLater();
            item.action->deleteLater();
            item.toolButton->deleteLater();
        }
    }
    group.rightSep->deleteLater();
    m_groups.erase(it);
}

int ButtonbarMgr::buttonPosition(QList<ButtonbarMgr::Button>& list, IWin* sw)
{
    for (int i = 0; i < list.count(); ++i) {
        ButtonbarMgr::Button& item = list[i];
        if (item.subwin->widget() == sw)
            return i;
    }
    return -1;
}

ButtonbarMgr::Button::Button(QMdiSubWindow* parent)
    : subwin(parent)
{
    const QString& buttonText = qobject_cast<IWin*>(parent->widget())->getButtonText();
    toolButton = new QToolButton(parent);
    toolButton->setCheckable(true);
    toolButton->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    toolButton->setIcon(parent->windowIcon());
    toolButton->setText(buttonText);
    toolButton->setToolTip(buttonText);

    menu = new QMenu(parent);
    menuHead = menu->addAction(buttonText);
    QFont f = menuHead->font();
    f.setBold(true);
    menuHead->setFont(f);
    menuHead->setDisabled(true);
    menuSep = menu->addSeparator();
    menuClose = menu->addAction(tr("Close"));
    connect(menuClose, &QAction::triggered, [parent](bool){
        parent->close();
    });
}
