/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (c) 2022  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "DCCQueryTableModel.h"
#include <stdexcept>

DCCQueryTableModel::DCCQueryTableModel(QObject* parent)
    : QAbstractTableModel(parent)
{
    m_items.reserve(20);
}

QModelIndex DCCQueryTableModel::index(int row, int column, const QModelIndex& parent) const
{
    return row < m_items.size() ? createIndex(row, column, nullptr)
                                : QModelIndex{};
}

int DCCQueryTableModel::rowCount(const QModelIndex& parent) const
{
    return static_cast<int>(m_items.size());
}

QVariant DCCQueryTableModel::data(const QModelIndex& index, int role) const
{
    if (role == Qt::DisplayRole) {
        const Item* item{};

        try {
            item = &(m_items.at(index.row()));
        }
        catch (const std::out_of_range&) {
            item = nullptr;
        }

        if (!item)
            return {};

        switch (index.column()) {
            case 0: return QString::fromStdString(item->sender.toString());
            case 1: return item->description;
            case 2: return "accept";
            case 3: return "decline";
        }
    }

    return {};
}

std::pair<QModelIndex,QModelIndex> DCCQueryTableModel::addQuery(const IRCPrefix& sender, const QString& description, const std::shared_ptr<DCC>& dcc)
{
    const auto pos = static_cast<int>(m_items.size());
    beginInsertRows(QModelIndex{}, pos, pos);
    m_items.emplace_back( Item{sender, description, dcc} );
    endInsertRows();

    return std::make_pair( createIndex(pos, 2, nullptr),
                           createIndex(pos, 3, nullptr)
                          );
}

void DCCQueryTableModel::removeAll()
{
    beginResetModel();
    m_items.clear();
    endResetModel();
}

void DCCQueryTableModel::removeEntry(const QModelIndex& index)
{
    if (!index.isValid() || (index.isValid() && index.row() > m_items.size()))
        return;

    beginRemoveRows(QModelIndex{}, index.row(), index.row());
    m_items.erase(m_items.begin() + index.row());
    endRemoveRows();
}

std::shared_ptr<DCC> DCCQueryTableModel::getDCCPtr(const QModelIndex& index)
{
    if (!index.isValid() || (index.isValid() && index.row() > m_items.size()))
        return nullptr;

    auto it = m_items.begin() + index.row();
    return it->dcc;
}

std::optional<IRCPrefix> DCCQueryTableModel::getSender(const QModelIndex& index)
{
    if (!index.isValid() || (index.isValid() && index.row() > m_items.size()))
        return {};

    auto it = m_items.begin() + index.row();
    return std::make_optional(it->sender);
}

Qt::ItemFlags DCCQueryTableModel::flags(const QModelIndex& index) const
{
    if (!index.isValid())
        return Qt::NoItemFlags;

    if (index.row() > 1)
        return Qt::ItemIsEnabled | Qt::ItemIsEditable;
    else
        return Qt::ItemIsEnabled;
}
