/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "IdealIRC.h"
#include "ui_IdealIRC.h"
#include "config.h"

#include "IWin/IWinStatus.h"
#include "IWin/IWinDCCChat.h"

#include <ConfigMgr.h>
#include <QDebug>
#include <QDir>
#include <QMessageBox>
#include <QScreen>
#include <QDesktopServices>

// Used externally. Initialized in the IdealIRC constructor.
IdealIRC* IIRC_MainWindow_Ptr{ nullptr };

IdealIRC::IdealIRC(QWidget* parent)
    : QMainWindow(parent)
    , ui(new Ui::IdealIRC)
    , confDlg(this)
    , aboutDlg(this)
    , dccQueryDlg(this)
{
    ui->setupUi(this);
    IIRC_MainWindow_Ptr = this;

    auto& conf = ConfigMgr::instance();
    const auto bbposcfg = conf.common("ButtonBarPosition");
    if (bbposcfg == "S") {
        addToolBar(Qt::BottomToolBarArea, ui->windowButtons);
    }
    else if (bbposcfg == "W") {
        addToolBar(Qt::LeftToolBarArea, ui->windowButtons);
    }
    else if (bbposcfg == "E") {
        addToolBar(Qt::RightToolBarArea, ui->windowButtons);
    }

    const QString title = QString("%1 %2")
                    .arg(windowTitle())
                    .arg(VERSION_STRING_NO_APPEND);
    setWindowTitle(title);
    trayIcon = new QSystemTrayIcon(QIcon(QStringLiteral(":/Icons/icon.png")), this);

    mdiManager = new MdiManager(*ui->mdiArea, *ui->windowButtons, *trayIcon);
    connect(mdiManager, &MdiManager::statusWindowCreated, this, &IdealIRC::onStatusWindowCreated);

    trayIcon->show();
    trayIcon->setToolTip(title);

    connect(trayIcon, &QSystemTrayIcon::activated, [this](QSystemTrayIcon::ActivationReason reason) {
        if (reason == QSystemTrayIcon::DoubleClick) {
            setVisible(!isVisible());
        }
    });

    trayIconMenu = new QMenu("IdealIRC", this);
    trayIconMenuShow = trayIconMenu->addAction("Show/Hide");
    trayIconMenu->addSeparator();
    trayIconMenuExit = trayIconMenu->addAction("Exit");

    connect(trayIconMenuShow, &QAction::triggered, [this](bool){
        setVisible(!isVisible());
    });

    connect(trayIconMenuExit, &QAction::triggered, [this](bool){
        close();
    });

    trayIcon->setContextMenu(trayIconMenu);

    connect(mdiManager, &MdiManager::readyForExit, this, &QMainWindow::close);
    connect(mdiManager, &MdiManager::subwindowSwitched, this, &IdealIRC::subwindowSwitched);
    connect(mdiManager, &MdiManager::connectionStateChange, this, &IdealIRC::updateConnectButtonState);
    connect(&confDlg, &IConfig::connectToServer, this, &IdealIRC::connectToServer);
    connect(&confDlg, &IConfig::disconnectFromServer, this, &IdealIRC::disconnectFromServer);
    setConnectButtonState(ConnectButtonState::Connect);

    IWin* statusw = mdiManager->createSubwindow(nullptr, "Status", IWin::Type::Status);
    statusw->refreshWindowTitle();

    QTimer::singleShot(10, [this]{ startup(); });
}

IdealIRC::~IdealIRC()
{
    delete mdiManager;
    delete trayIcon;
    delete ui;
}

void IdealIRC::saveSubwindowGeometries()
{
    for (auto* mdiwin : mdiManager->mdiChildren()) {
        IWin* subwin = dynamic_cast<IWin*>(mdiwin->widget());
        if ((subwin->getType() == IWin::Type::Channel || subwin->getType() == IWin::Type::Private)
                && !mdiwin->isMaximized())
        {

            QString network;
            try {
                const auto* status = dynamic_cast<IWinStatus*>(subwin->getStatusParent());
                network = QString::fromStdString( status->getConnection().isupport().at("NETWORK") );
            }
            catch (...) {}

            QString persistentWindowName;
            if (network.isEmpty())
                persistentWindowName = subwin->getButtonText();
            else
                persistentWindowName = QStringLiteral("%1_%2")
                        .arg(network)
                        .arg(subwin->getButtonText());

            const auto pos = mdiwin->pos();
            const auto size = mdiwin->size();
            const auto geo = QStringLiteral("%1,%2,%3,%4")
                    .arg(pos.x())
                    .arg(pos.y())
                    .arg(size.width())
                    .arg(size.height());

            auto& conf = ConfigMgr::instance();
            conf.setWindow(persistentWindowName, geo);
        }
    }
}

void IdealIRC::closeEvent(QCloseEvent* evt)
{
    ConfigMgr& conf = ConfigMgr::instance();

    /*
     * Store geometry
    */
    if (!isMaximized()) {
        auto screen = QApplication::screenAt(pos());
        auto screenGeo = screen->geometry();
        auto X = QString::number(x() - screenGeo.x());
        auto Y = QString::number(y() - screenGeo.y());
        auto W = QString::number(width());
        auto H = QString::number(height());
        conf.setGeometry("X", X);
        conf.setGeometry("Y", Y);
        conf.setGeometry("Width", W);
        conf.setGeometry("Height", H);
        conf.setGeometry("Maximized", "0");
    }
    else {
        conf.setGeometry("Maximized", "1");
    }

    /*
     * Buttonbar placement
    */
    {
        auto bbpos = toolBarArea(ui->windowButtons);
        QString bbposcfg;
        switch (bbpos) {
            case Qt::LeftToolBarArea:
                bbposcfg = "W";
                break;
            case Qt::RightToolBarArea:
                bbposcfg = "E";
                break;
            case Qt::BottomToolBarArea:
                bbposcfg = "S";
                break;
            default:
                bbposcfg = "N";
        }
        conf.setCommon("ButtonBarPosition", bbposcfg);
    }

    /*
     * Initialize closing
    */
    int serversWaiting = mdiManager->connectionsOnlineCount();
    if (serversWaiting > 0) {
        auto btn = QMessageBox::question(this, tr("Close IdealIRC"), tr("There are %1 connections active.\nDo you really want to exit IdealIRC?").arg(serversWaiting));
        if (btn == QMessageBox::Yes)
            mdiManager->broadcastProgramExit();
        evt->ignore();
    }
    else {
        saveSubwindowGeometries();
        ConfigMgr::instance().save();
        ScriptManager::event("exit");
        ScriptManager::instance()->unloadAllScripts();
        emit aboutToClose();
        evt->accept();
    }
}

void IdealIRC::startup()
{
    const ConfigMgr& conf = ConfigMgr::instance();
    if (conf.common("ShowOptions") == "1")
        on_actionOptions_triggered();

    const auto geoX = conf.geometry("X").toInt();
    const auto geoY = conf.geometry("Y").toInt();
    const auto geoWidth = conf.geometry("Width").toInt();
    const auto geoHeight = conf.geometry("Height").toInt();

    auto geo = geometry();
    auto screen = QApplication::screenAt(QCursor::pos());
    auto screenGeo = screen->geometry();

    if (geoX != -1 && geoY != -1) {
        geo = {
                geoX + screenGeo.x(),
                geoY + screenGeo.y(),
                geoWidth,
                geoHeight
              };
    }
    else {
        int x = screenGeo.width() / 2;
        int y = screenGeo.height() / 2;
        x -= geoWidth / 2;
        y -= geoHeight / 2;
        geo = {
                x + screenGeo.x(),
                y + screenGeo.y(),
                geoWidth,
                geoHeight
              };
    }

    setGeometry(geo);

    if (conf.geometry("Maximized") == "1")
        showMaximized();

    ui->actionHide_toolbar->setChecked( conf.common("ToolbarHidden") == "1" );
    if (ui->actionHide_toolbar->isChecked())
        ui->toolBar->hide();

    ScriptManager::init(mdiManager, this);
    scriptEvent("start");
}

void IdealIRC::subwindowSwitched()
{
    updateConnectButtonState();
    ScriptManager::setContext(mdiManager->currentStatus());
}

void IdealIRC::updateConnectButtonState()
{
    IWinStatus* status = mdiManager->currentStatus();
    bool isConnected = status->getConnection().isConnected();
    setConnectButtonState(isConnected ? ConnectButtonState::Disconnect : ConnectButtonState::Connect);
}

void IdealIRC::setConnectButtonState(IdealIRC::ConnectButtonState state)
{
    switch (state) {
    case ConnectButtonState::Connect:
        ui->actionConnect->setText(tr("Connect"));
        ui->actionConnect->setIcon( QIcon(":/Icons/connect.png") );
        break;
    case ConnectButtonState::Disconnect:
        ui->actionConnect->setText(tr("Disconnect"));
        ui->actionConnect->setIcon( QIcon(":/Icons/disconnect.png") );
        break;
    }
    ui->actionConnect->setData(static_cast<int>(state));
}

void IdealIRC::connectToServer(bool newStatus)
{
    IWinStatus* status{ nullptr };
    if (newStatus)
        status = static_cast<IWinStatus*>(mdiManager->createSubwindow(nullptr, "Status", IWin::Type::Status));
    else
        status = mdiManager->currentStatus();

    const ConfigMgr& conf = ConfigMgr::instance();
    QString server = conf.connection("Server");
    QString nickname = conf.connection("Nickname");
    QString realname = conf.connection("Realname");
    QString username = conf.connection("Username");
    QString password = conf.connection("Password");
    QString sasl = conf.connection("SASL");

    if (server.isEmpty() || nickname.isEmpty() || realname.isEmpty() || username.isEmpty()) {
        confDlg.show();
        return;
    }

    bool ssl = conf.connection("SSL").toInt();
    QString host;
    quint16 port;
    if (server.contains(':')) {
        QString strport = server.split(':')[1];
        port = strport.toUShort();
        server = server.remove(":"+strport);
    }
    else
        port = 6667;
    host = server;

    IRC& con = status->getConnection();
    con.setHostname(host.toStdString(), ssl);
    con.setPort(std::to_string(port));

    con.setRealname(realname.toStdString());
    con.setIdent(username.toStdString());
    con.setNickname(nickname.toStdString());
    con.setPassword(password.toStdString());
    con.setV3SASL(sasl.toStdString());

    bool manualKeepaliveEnabled = conf.common("ManualKeepaliveEnabled").toInt();
    int manualKeepalive = manualKeepaliveEnabled ? conf.common("ManualKeepalive").toInt() : 0;
    con.setManualKeepalive(std::chrono::seconds(manualKeepalive));

    const auto cfgSSLExcept_Expired = conf.common("SSLExpired");
    const auto cfgSSLExcept_Selfsigned = conf.common("SSLSelfsigned");
    const auto cfgSSLExcept_CNMismatch = conf.common("SSLCNMismatch");
    con.exceptSSL_Expired(cfgSSLExcept_Expired == "1");
    con.exceptSSL_SelfSigned(cfgSSLExcept_Selfsigned == "1");
    con.exceptSSL_CNMismatch(cfgSSLExcept_CNMismatch == "1");

    if (con.isConnected()) {
        auto quitMsg = conf.common("QuitMessage").toStdString();
        if (con.disconnectFromServer(quitMsg) == IRCError::NotConnected)
            qWarning() << "Trying to disconnect from an already closed connection.";

        auto e = con.tryConnect();
        if (e != IRCError::NoError)
            status->print(PrintType::ProgramInfo, IRCErrorToString(e).c_str());
    }
    else {
        setConnectButtonState(ConnectButtonState::Disconnect);
        auto e = con.tryConnect();
        if (e != IRCError::NoError)
            status->print(PrintType::ProgramInfo, IRCErrorToString(e).c_str());
    }
}

void IdealIRC::disconnectFromServer()
{
    const ConfigMgr& conf = ConfigMgr::instance();

    IWinStatus* status = mdiManager->currentStatus();
    auto& connection = status->getConnection();
    connection.expectDisconnect();

    auto quitMsg = conf.common("QuitMessage").toStdString();
    if (connection.disconnectFromServer(quitMsg) == IRCError::NotConnected)
        qWarning() << "Trying to disconnect from an already closed connection.";
}

void IdealIRC::onStatusWindowCreated(IWinStatus* window)
{
    connect(&(window->getConnection()), &IRC::showTrayMessage, [this] (const QString& title, const QString& message) {
        if (!isVisible() || !isActiveWindow()) {
            auto& conf = ConfigMgr::instance();
            trayIcon->showMessage(title, message, QSystemTrayIcon::Information, 1000 * conf.common("TrayNotifyDelay").toInt());
        }
    });

    connect(&(window->getConnection()), &IRC::DCCRequested, [this,window](const std::shared_ptr<DCC>& dcc, const IRCPrefix& sender, const QString& type, const QString& message) {
        onDCCRequested(*window, dcc, sender, type, message);
    });
}

void IdealIRC::onDCCRequested(const IWinStatus& statusParent, const std::shared_ptr<DCC>& dcc, const IRCPrefix& sender, const QString& type, const QString& message)
{
    /*
     * Need to check any DCC windows already open in the given status parent of the inbound DCC for pending reverses.
     * Then compare for any matching pending reverses with this inbound DCC.
    */
    auto dccChatList = MdiManager::instance().childrenOf(&statusParent, IWin::Type::DCCChat);
    bool requestHandled{ false };
    for (auto* subwin : dccChatList) {
        auto* dccwin = dynamic_cast<IWinDCCChat*>(subwin);
        if (!dccwin) {
            qWarning() << "Unable to cast a window marked as type DCCChat!?";
            continue;
        }

        if (QString::fromStdString(sender.toString()) == dccwin->getButtonText() && dccwin->isReverseInitiatePending()) {
            requestHandled = true;
            dccwin->initiateReversed(dcc);
        }
    }

    if (!requestHandled) {
        dccQueryDlg.addQuery(sender, type, dcc);
        dccQueryDlg.show();
    }
}

void IdealIRC::on_actionConnect_triggered()
{
    ConnectButtonState state = static_cast<ConnectButtonState>(ui->actionConnect->data().toInt());
    if (state == ConnectButtonState::Connect)
        connectToServer(false);

    else if (state == ConnectButtonState::Disconnect) {
        IRC& connection = mdiManager->currentStatus()->getConnection();
        const ConfigMgr& conf = ConfigMgr::instance();
        if (!connection.isOnline())
            setConnectButtonState(ConnectButtonState::Connect);

        connection.expectDisconnect();
        auto e = connection.disconnectFromServer(conf.common("QuitMessage").toStdString());
        if (e == IRCError::NotConnected)
               qWarning() << "Trying to disconnect from an already closed connection.";
    }
}

void IdealIRC::on_actionOptions_triggered()
{
    confDlg.show();
    if (mdiManager->currentStatus()->getConnection().isConnected())
        confDlg.showDisconnectButton();
    else
        confDlg.hideDisconnectButton();
}

void IdealIRC::on_actionAbout_IdealIRC_triggered()
{
    aboutDlg.show();
}

void IdealIRC::on_actionScripts_triggered()
{
    ScriptManager::instance()->show();
}

void IdealIRC::on_actionOnline_wiki_triggered()
{
    QDesktopServices::openUrl( QUrl{"http://wiki.idealirc.org/"} );
}

void IdealIRC::on_actionHide_toolbar_triggered()
{
    ConfigMgr& conf = ConfigMgr::instance();

    conf.setCommon("ToolbarHidden", ui->actionHide_toolbar->isChecked() ? "1" : "0");
    ui->toolBar->setVisible(!ui->actionHide_toolbar->isChecked());
}

void IdealIRC::on_actionDCC_queries_triggered()
{
    dccQueryDlg.show();
}
