/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "IniFile.h"
#include <vector>
#include <algorithm>
#include <fstream>

namespace {
/*
 * Not unicode friendly.
 * Used to make case-insensitive comparisons of keys and section names.
*/
bool strEquals(const std::string& l, const std::string& r)
{
    if (l.size() != r.size())
        return false;
    
    return std::equal(l.begin(), l.end(), r.begin(),
        [](char cl, char cr) {
            return toupper(cl) == toupper(cr);
        }
    );
}

void trimFrontAndBack(std::string& str) {
    while (!str.empty() && std::isspace(str.front()))
        str.erase(str.begin());
    while (!str.empty() && std::isspace(str.back()))
        str.erase(str.end() - 1);
}

}

struct Item
{
    std::string key;
    std::string val;
};

struct Section
{
    std::string name;
    std::vector<Item> items;
};

struct IniFilePriv
{
    IniFile::Error setError(IniFile::Error e)
    {
        lastError = e;
        return e;
    }

    IniFile::Error lastError{ IniFile::Error::NoError };
    std::string filename;
    std::vector<Section> sections;
};

IniFile::IniFile(const std::string& filename)
    : mp(std::make_unique<IniFilePriv>())
{
    mp->filename = filename;
    reload();
}

IniFile::IniFile(IniFile&& other) noexcept
    : mp(std::move(other.mp))
{
}

IniFile::~IniFile() = default;

IniFile& IniFile::operator=(IniFile&& other) noexcept
{
    mp = std::move(other.mp);
    return *this;
}

IniFile::Error IniFile::error() const
{
    return mp->lastError;
}

std::size_t IniFile::count() const
{
    return mp->sections.size();
}

std::size_t IniFile::count(const std::string& section) const
{
    auto it = std::find_if(mp->sections.begin(), mp->sections.end(),
        [section](const Section& sn){
            return strEquals(sn.name, section);
        }
    );

    return it != mp->sections.end()
            ? it->items.size() : 0;
}

std::string IniFile::read(const std::string& section, const std::string& item, const std::string& defaultValue) const
{
    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
        [section](const Section& sn){
            return strEquals(sn.name, section);
        }
    );

    if (s_it == mp->sections.end())
        return defaultValue;

    const auto& items = s_it->items;
    auto i_it = std::find_if(items.begin(), items.end(),
        [item](const Item& i){
           return strEquals(i.key, item);
        }
    );

    if (i_it != items.end())
        return i_it->val;
    else
        return defaultValue;
}

std::string IniFile::read(const std::string& section, int itemPos, const std::string& defaultValue) const
{
    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    if (s_it == mp->sections.end())
        return defaultValue;

    try {
        const auto& items = s_it->items;
        return items.at(itemPos).val;
    }
    catch (const std::out_of_range&) {
        return defaultValue;
    }
}

std::string IniFile::read(int sectionPos, const std::string& item, const std::string& defaultValue) const
{
    try {
        const auto& items = mp->sections[sectionPos].items;
        auto i_it = std::find_if(items.begin(), items.end(),
             [item](const Item& i){
                 return strEquals(i.key, item);
             }
        );

        if (i_it != items.end())
            return i_it->val;
        else
            return defaultValue;
    }
    catch (const std::out_of_range&) {
        return defaultValue;
    }
}

std::string IniFile::read(int sectionPos, int itemPos, const std::string& defaultValue) const
{
    try {
        const auto& items = mp->sections[sectionPos].items;
        return items.at(itemPos).val;
    }
    catch (const std::out_of_range&) {
        return defaultValue;
    }
}

std::string IniFile::item(const std::string& section, int itemPos) const
{
    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    if (s_it == mp->sections.end())
        return {};

    try {
        const auto& items = s_it->items;
        return items.at(itemPos).key;
    }
    catch (const std::out_of_range&) {
        return {};
    }
}

std::string IniFile::item(int sectionPos, int itemPos) const
{
    try {
        const auto& items = mp->sections[sectionPos].items;
        return items.at(itemPos).key;
    }
    catch (const std::out_of_range&) {
        return {};
    }
}

std::string IniFile::section(int sectionPos) const
{
    try {
        const auto& section = mp->sections[sectionPos];
        return section.name;
    }
    catch (const std::out_of_range&) {
        return {};
    }
}

void IniFile::write(const std::string& section)
{
    if (exist(section))
        return;

    auto& s = mp->sections.emplace_back();
    s.name = section;
}

void IniFile::write(const std::string& section, const std::string& item, const std::string& value)
{
    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    if (s_it == mp->sections.end()) {
        mp->sections.emplace_back();
        s_it = mp->sections.end() - 1;
        s_it->name = section;
    }

    auto& items = s_it->items;
    auto i_it = std::find_if(items.begin(), items.end(),
         [item](const Item& i){
             return strEquals(i.key, item);
         }
    );

    if (i_it == items.end()) {
        s_it->items.emplace_back();
        i_it = s_it->items.end() - 1;
        i_it->key = item;
    }

    i_it->val = value;
}

IniFile::Error IniFile::rename(const std::string& section, const std::string& newSection)
{
    if (exist(newSection))
        return mp->setError(IniFile::Error::NewAlreadyExists);

    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    if (s_it == mp->sections.end()) {
        return mp->setError(IniFile::Error::OldDoesntExist);
    }
    else {
        s_it->name = newSection;
    }

    return mp->setError(IniFile::Error::NoError);
}

IniFile::Error IniFile::rename(const std::string& section, const std::string& item, const std::string& newItem)
{
    if (exist(section, newItem))
        return mp->setError(IniFile::Error::NewAlreadyExists);

    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    if (s_it == mp->sections.end())
        return mp->setError(IniFile::Error::SectionDontExist);

    auto& items = s_it->items;
    auto i_it = std::find_if(items.begin(), items.end(),
         [item](const Item& i){
             return strEquals(i.key, item);
         }
    );

    if (i_it == items.end()) {
        return mp->setError(IniFile::Error::OldDoesntExist);
    }
    else {
        i_it->key = newItem;
    }

    return mp->setError(IniFile::Error::NoError);
}

IniFile::Error IniFile::remove(const std::string& section)
{
    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    if (s_it == mp->sections.end()) {
        return mp->setError(IniFile::Error::SectionDontExist);
    }
    else {
        mp->sections.erase(s_it);
    }

    return mp->setError(IniFile::Error::NoError);
}

IniFile::Error IniFile::remove(const std::string& section, const std::string& item)
{
    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    if (s_it == mp->sections.end())
        return mp->setError(IniFile::Error::SectionDontExist);

    auto& items = s_it->items;
    auto i_it = std::find_if(items.begin(), items.end(),
         [item](const Item& i){
             return strEquals(i.key, item);
         }
    );

    if (i_it == items.end()) {
        return mp->setError(IniFile::Error::ItemDontExist);
    }
    else {
        items.erase(i_it);
    }

    return mp->setError(IniFile::Error::NoError);
}

bool IniFile::exist(const std::string& section) const
{
    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    return s_it != mp->sections.end();
}

bool IniFile::exist(const std::string& section, const std::string& item) const
{
    auto s_it = std::find_if(mp->sections.begin(), mp->sections.end(),
         [section](const Section& sn){
             return strEquals(sn.name, section);
         }
    );

    if (s_it == mp->sections.end())
        return false;

    auto& items = s_it->items;
    auto i_it = std::find_if(items.begin(), items.end(),
         [item](const Item& i){
             return strEquals(i.key, item);
         }
    );

    return i_it != items.end();
}

IniFile::Error IniFile::flush()
{
#if defined(_WIN32) || defined(_WIN64)
    constexpr auto EOL = "\r\n";
#else
    constexpr auto EOL = "\n";
#endif

    std::string data;
    data.reserve(1024);
    for (const auto& section : mp->sections) {
        if (!data.empty())
            data += EOL;
        data += '[';
        data += section.name;
        data += ']';
        data += EOL;
        for (const auto& item : section.items) {
            data += item.key;
            data += '=';
            data += item.val;
            data += EOL;
        }
    }

    std::ofstream file(mp->filename);
    if (!file.is_open())
        return mp->setError(IniFile::Error::CannotWriteFile);
    file << data;
    file.close();
    return mp->setError(IniFile::Error::NoError);
}

IniFile::Error IniFile::reload()
{
    std::ifstream file(mp->filename);
    if (!file.is_open())
        return mp->setError(IniFile::Error::CannotOpenFile);

    std::string section;
    std::string line;
    while (std::getline(file, line)) {
        if (line.back() == 0x0D)
            line.pop_back(); // Windows ending CRLF, CR is read, but we drop it.

        trimFrontAndBack(line);

        if (line.empty())
            continue;

        /* Sections */
        if (line.front() == '[' && line.back() == ']') {
            section = line.substr(1, line.size() - 2);
        }

        /* Items */
        else if (auto p = line.find('='); p != std::string::npos && !section.empty()) {
            const auto item = line.substr(0, p);
            const auto value = line.substr(p + 1);
            write(section, item, value);
        }

        /* Garbage; malformed items or items outside of section */
        else {
        }
    }
    file.close();

    return mp->setError(IniFile::Error::NoError);
}

