/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "DialogUtils.h"
#include "Script/ValueExtract.h"
#include "Script/ScriptException.h"
#include "Error.h"
#include "Script/Script.h"

using namespace Builtin::Error;

namespace Builtin::DialogUtils {

ValueHolder dialog(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("dialog", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("dialog", 1);

    return script.dialogConstruct(ValueExtract(args[0]).toString());
}

ValueHolder dlgvalidhndl(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("dlgvalidhndl", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::Integer)
        return false;

    return script.dialogIsValidHandle(ValueExtract(args[0]).toInt());
}

ValueHolder dlgclose(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("dlgclose", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlgclose", 1);

    script.dialogClose(ValueExtract(args[0]).toInt());
    return ValueHolder();
}

ValueHolder dlghide(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("dlghide", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlghide", 1);

    script.dialogHide(ValueExtract(args[0]).toInt());
    return ValueHolder();
}

ValueHolder dlgshow(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("dlgshow", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlgshow", 1);

    script.dialogShow(ValueExtract(args[0]).toInt());
    return ValueHolder();
}

ValueHolder dlglist(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("dlglist", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("dlglist", 1);

    return script.dialogList(ValueExtract(args[0]).toString());
}

ValueHolder dlgsetattr(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 3)
        throwInsufficientParameters("dlgsetattr", args.size(), 3);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlgsetattr", 1);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("dlgsetattr", 2);

    int handle = ValueExtract(args[0]).toInt();
    const auto& attribute = ValueExtract(args[1]).toString();
    ValueHolder& value = args[2];
    script.dialogSetAttr(handle, attribute, value);

    return ValueHolder();
}

ValueHolder dlggetattr(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("dlggetattr", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlggetattr", 1);

    int handle = ValueExtract(args[0]).toInt();
    const auto& attribute = ValueExtract(args[1]).toString();

    return script.dialogGetAttr(handle, attribute);
}

ValueHolder dlgsetwidgetattr(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 4)
        throwInsufficientParameters("dlgsetwidgetattr", args.size(), 4);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlgsetwidgetattr", 1);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("dlgsetwidgetattr", 2);

    if (args[2].getType() != ValueHolder::Type::String)
        throwNotAString("dlgsetwidgetattr", 3);

    int handle = ValueExtract(args[0]).toInt();
    const auto& widgetName = ValueExtract(args[1]).toString();
    const auto& attribute = ValueExtract(args[2]).toString();
    ValueHolder& value = args[3];
    script.dialogSetWidgetAttr(handle, widgetName, attribute, value);

    return ValueHolder();
}

ValueHolder dlggetwidgetattr(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 3)
        throwInsufficientParameters("dlggetwidgetattr", args.size(), 3);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlggetwidgetattr", 1);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("dlgsetwidgetattr", 2);

    int handle = ValueExtract(args[0]).toInt();
    const auto& widgetName = ValueExtract(args[1]).toString();
    const auto& attribute = ValueExtract(args[2]).toString();

    return script.dialogGetWidgetAttr(handle, widgetName, attribute);
}

ValueHolder dlgtablecount(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("dlgtablecount", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlgtablecount", 1);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("dlgtablecount", 2);

    int handle = ValueExtract(args[0]).toInt();
    const auto& widgetName = ValueExtract(args[1]).toString();

    return script.dialogTableWidgetCount(handle, widgetName);
}

ValueHolder dlgtableinsert(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 3)
        throwInsufficientParameters("dlgtableinsert", args.size(), 3);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlgtableinsert", 1);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("dlgtableinsert", 2);

    if (args[2].getType() != ValueHolder::Type::Map)
        throwNotAMap("dlgtableinsert", 3);

    int handle = ValueExtract(args[0]).toInt();
    const auto& widgetName = ValueExtract(args[1]).toString();
    const auto& cols = ValueExtract(args[2]).toMap();

    script.dialogTableWidgetInsert(handle, widgetName, cols);
    return ValueHolder();
}

ValueHolder dlgtableremove(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 3)
        throwInsufficientParameters("dlgtableremove", args.size(), 3);

    if (args[0].getType() != ValueHolder::Type::Integer)
        throwNotADialogHandle("dlgtableremove", 1);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("dlgtableremove", 2);

    if (args[2].getType() != ValueHolder::Type::Integer)
        throwNotANumber("dlgtableremove", 3);

    int handle = ValueExtract(args[0]).toInt();
    const auto& widgetName = ValueExtract(args[1]).toString();
    const auto& index = ValueExtract(args[2]).toInt();

    script.dialogTableWidgetRemove(handle, widgetName, index);
    return ValueHolder();
}

ValueHolder dlgtableselected(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("dlgtableselected", args.size(), 2);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("dlgtableselected", 2);

    int handle = ValueExtract(args[0]).toInt();
    const auto& widgetName = ValueExtract(args[1]).toString();

    return script.dialogTableWidgetSelected(handle, widgetName);
}

ValueHolder dlgtablerow(Script& script, std::vector<ValueHolder>& args)
{
    if (args.size() != 3)
        throwInsufficientParameters("dlgtablerow", args.size(), 3);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("dlgtablerow", 2);

    if (args[2].getType() != ValueHolder::Type::Integer)
        throwNotANumber("dlgtablerow", 3);

    int handle = ValueExtract(args[0]).toInt();
    const std::string& widgetName = ValueExtract(args[1]).toString();
    int row = ValueExtract(args[2]).toInt();

    return script.dialogTableWidgetRow(handle, widgetName, row);
}

}
