/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "GeneralUtils.h"
#include "Script/ValueExtract.h"
#include "Script/ScriptException.h"
#include "Error.h"
#include <locale>
#include <iostream>

using namespace Builtin::Error;

namespace {
inline std::string decimalSeparator()
{
    std::string ret;
    ret += std::use_facet<std::numpunct<char>>(std::cout.getloc()).decimal_point();
    return ret;
}
}

namespace Builtin::GeneralUtils {

ValueHolder toint(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("toint", args.size(), 1);

    using VType = ValueHolder::Type;
    switch (args[0].getType()) {
    case VType::Void:
        throw ScriptException("toint: A void value cannot convert to int");
    case VType::Map:
        throw ScriptException("toint: A map cannot convert to int");
    case VType::Integer: [[fallthrough]];
    case VType::Real: [[fallthrough]];
    case VType::Bool:
        return ValueExtract(args[0]).toInt();
    case VType::String:
    {
        const std::string& val = ValueExtract(args[0]).toString();
        if (val == "false")
            return 0;
        if (val == "true")
            return 1;
        return std::stoi(val);
    }
    }

    return {}; // Compiler complains even though all cases are covered in the above switch statement.
}

ValueHolder tostr(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("tostr", args.size(), 1);

    return args[0].toStdString();
}

ValueHolder toreal(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("toreal", args.size(), 1);

    using VType = ValueHolder::Type;
    switch (args[0].getType()) {
    case VType::Void:
        throw ScriptException("toint: A void value cannot convert to real");
    case VType::Map:
        throw ScriptException("toint: A map cannot convert to double");
    case VType::Integer: [[fallthrough]];
    case VType::Real: [[fallthrough]];
    case VType::Bool:
        return ValueExtract(args[0]).toReal();
    case VType::String:
    {
        std::string val = ValueExtract(args[0]).toString();
        if (val == "false")
            return 0.0;
        if (val == "true")
            return 1.0;
        auto dotpos = val.find('.');
        if (dotpos != val.npos)
            val.replace(dotpos, 1, decimalSeparator());
        return std::stod(val);
    }
    }

    return {}; // Compiler complains even though all cases are covered in the above switch statement.
}

ValueHolder tobool(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("tobool", args.size(), 1);

    using VType = ValueHolder::Type;
    switch (args[0].getType()) {
    case VType::Void:
        throw ScriptException("toint: A void value cannot convert to bool");
    case VType::Map:
        throw ScriptException("toint: A map cannot convert to bool");
    case VType::Integer: [[fallthrough]];
    case VType::Real: [[fallthrough]];
    case VType::Bool:
        return ValueExtract(args[0]).toBool();
    case VType::String:
    {
        std::string val = ValueExtract(args[0]).toString();
        if (val == "true" || val == "1")
            return true;
        else
            return false;
    }
    }

    return {}; // Compiler complains even though all cases are covered in the above switch statement.
}

ValueHolder isvoid(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("isvoid", args.size(), 1);

    return args[0].getType() == ValueHolder::Type::Void;
}

ValueHolder isint(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("isint", args.size(), 1);

    return args[0].getType() == ValueHolder::Type::Integer;
}

ValueHolder isreal(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("isreal", args.size(), 1);

    return args[0].getType() == ValueHolder::Type::Real;
}

ValueHolder isbool(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("isbool", args.size(), 1);

    return args[0].getType() == ValueHolder::Type::Bool;
}

ValueHolder isstr(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("isstring", args.size(), 1);

    return args[0].getType() == ValueHolder::Type::String;
}

ValueHolder ismap(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("ismap", args.size(), 1);

    return args[0].getType() == ValueHolder::Type::Map;
}

ValueHolder decimalsep(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 0)
        throwInsufficientParameters("decimalsep", 0, 1);
    return decimalSeparator();
}

}
