/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "ListUtils.h"
#include "Error.h"
#include "Script/ValueExtract.h"
#include "Script/ScriptException.h"
#include <stdexcept>

const std::string LIST_END_MAGIC = "_list_end_";

using namespace Builtin::Error;

namespace Builtin::ListUtils {

ValueHolder list(Script&, std::vector<ValueHolder>& args)
{
    ValueArray ret;
    int idx = 0;
    for (auto& val : args)
        ret.emplace(std::to_string(idx++), new ValueHolder(std::move(val)));
    ret.emplace(LIST_END_MAGIC, new ValueHolder(idx));
    return ValueHolder(std::move(ret));
}

ValueHolder listpush(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("listpush", args.size(), 2);

    if (!args[0].isReference())
        throwNotAReference("listpush", 1);

    if (args[0].getType() != ValueHolder::Type::Map)
        throwNotAList("listpush", 1);

    try {
        ValueArray& arr = ValueExtract(args[0]).toMap();
        ValueHolder& endVal = *arr[LIST_END_MAGIC];
        arr.emplace(std::to_string(ValueExtract(endVal).toInt()), new ValueHolder(args[1]));
        endVal = endVal + 1;
    }  catch (const std::out_of_range&) {
        throwNotAList("listpush", 1);
    }
    return ValueHolder();
}

ValueHolder listpop(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("listpop", args.size(), 1);

    if (!args[0].isReference())
        throwNotAReference("listpop", 1);

    if (args[0].getType() != ValueHolder::Type::Map)
        throwNotAList("listpop", 1);

    try {
        ValueArray& arr = ValueExtract(args[0]).toMap();
        ValueHolder& endVal = *arr[LIST_END_MAGIC];
        int end = ValueExtract(endVal).toInt();
        if (arr.size() == 0)
            throw ScriptException("Cannot pop an empty list");
        ValueHolder ret = *arr.at(std::to_string(end - 1));
        arr.erase(std::to_string(end - 1));
        endVal = end - 1;
        return ret;
    }  catch (const std::out_of_range&) {
        throwNotAList("listpop", 1);
    }
}

ValueHolder listlen(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("listlen", args.size(), 1);

    if (!args[0].isReference())
        throwNotAReference("listlen", 1);

    if (args[0].getType() != ValueHolder::Type::Map)
        throwNotAList("listlen", 1);

    ValueArray& arr = ValueExtract(args[0]).toMap();
    if (arr.count(LIST_END_MAGIC) < 1)
        throwNotAList("listlen", 1);

    return *arr[LIST_END_MAGIC];
}

ValueHolder listremove(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("listremove", args.size(), 2);

    if (!args[0].isReference())
        throwNotAReference("listremove", 1);

    if (args[0].getType() != ValueHolder::Type::Map)
        throwNotAList("listremove", 1);

    if (args[1].getType() != ValueHolder::Type::Integer)
        throwNotAWholeNumber("listremove", 2);

    int index = ValueExtract(args[1]).toInt();
    ValueArray& arr = ValueExtract(args[0]).toMap();
    ValueHolder& endVal = *arr[LIST_END_MAGIC];
    int end = ValueExtract(endVal).toInt();

    if (index < 0 || index >= end)
        throwIndexOutOfRange("listremove", 2);

    arr.erase(std::to_string(index));
    for (int i = index + 1; i < end; ++i) {
        auto node = arr.extract(std::to_string(i));
        node.key() = std::to_string(i - 1);
        arr.insert(std::move(node));
    }
    endVal = end - 1;

    return ValueHolder();
}

ValueHolder listjoin(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("listjoin", args.size(), 2);

    if (!args[0].isReference())
        throwNotAReference("listjoin", 1);

    if (args[0].getType() != ValueHolder::Type::Map)
        throwNotAList("listjoin", 1);

    if (!args[1].isReference())
        throwNotAReference("listjoin", 2);

    if (args[1].getType() != ValueHolder::Type::Map)
        throwNotAList("listjoin", 2);

    ValueArray& M1arr = ValueExtract(args[0]).toMap();
    ValueHolder& M1endVal = *M1arr[LIST_END_MAGIC];
    int M1end = ValueExtract(M1endVal).toInt();

    ValueArray& M2arr = ValueExtract(args[1]).toMap();
    ValueHolder& M2endVal = *M2arr[LIST_END_MAGIC];
    int M2end = ValueExtract(M2endVal).toInt();

    for (int i = 0; i < M2end; ++i) {
        const int newIdx = i + M1end;
        auto& vh = *M2arr[std::to_string(i)];
        M1arr.emplace(std::to_string(newIdx), new ValueHolder(vh));
    }
    M1endVal = M1end + M2end;
    return ValueHolder();
}

ValueHolder listsplice(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("listsplice", args.size(), 2);

    if (!args[0].isReference())
        throwNotAReference("listsplice", 1);

    if (args[0].getType() != ValueHolder::Type::Map)
        throwNotAList("listsplice", 1);

    if (!args[1].isReference())
        throwNotAReference("listsplice", 2);

    if (args[1].getType() != ValueHolder::Type::Map)
        throwNotAList("listsplice", 2);

    ValueArray ret;

    ValueArray& M1arr = ValueExtract(args[0]).toMap();
    ValueHolder& M1endVal = *M1arr[LIST_END_MAGIC];
    int M1end = ValueExtract(M1endVal).toInt();

    for (int i = 0; i < M1end; ++i) {
        auto& vh = *M1arr[std::to_string(i)];
        ret.emplace(std::to_string(i), new ValueHolder(vh));
    }

    ValueArray& M2arr = ValueExtract(args[1]).toMap();
    ValueHolder& M2endVal = *M2arr[LIST_END_MAGIC];
    int M2end = ValueExtract(M2endVal).toInt();

    for (int i = 0; i < M2end; ++i) {
        const int newIdx = i + M1end;
        auto& vh = *M2arr[std::to_string(i)];
        ret.emplace(std::to_string(newIdx), new ValueHolder(vh));
    }
    ret.emplace(LIST_END_MAGIC, new ValueHolder(M1end + M2end));

    return ValueHolder(std::move(ret));
}

}
