/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "MapUtils.h"
#include "Error.h"
#include "Script/ValueExtract.h"
#include "Script/ScriptException.h"

#include <iostream>

using namespace Builtin::Error;

namespace Builtin::MapUtils {

ValueHolder map(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 0)
        throwInsufficientParameters("map", args.size(), 0);
    return ValueArray();
}

ValueHolder maplen(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("maplen", args.size(), 1);

    const ValueArray& arr = ValueExtract(args[0]).toMap();
    return static_cast<int>(arr.size());
}

ValueHolder mapremove(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("mapremove", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::Map)
        throwNotAMap("mapremove", 1);

    if (args[1].getType() == ValueHolder::Type::Void
        || args[1].getType() == ValueHolder::Type::Map)
        throw ScriptException("mapremove: 2nd parameter must either be an integer, real, bool or string");

    ValueArray& arr = ValueExtract(args[0]).toMap();
    return arr.erase(args[1].toStdString()) > 0;
}

ValueHolder mapdump(Script&, std::vector<ValueHolder>& args)
{
    std::cout << "--- Map dump ---\n";
    ValueArray& arr = ValueExtract(args[0]).toMap();
    for (const auto& item : arr) {
        auto type = item.second->getType();
        switch (type) {
        case ValueHolder::Type::Void:
            std::cout << '"' << item.first << "\" = Void\n";
            break;
        case ValueHolder::Type::Integer:
            std::cout << '"' << item.first << "\" = Integer(" << ValueExtract(*item.second).toInt() << ")\n";
            break;
        case ValueHolder::Type::Real:
            std::cout << '"' << item.first << "\" = Real(" << ValueExtract(*item.second).toReal() << ")\n";
            break;
        case ValueHolder::Type::Bool:
            std::cout << '"' << item.first << "\" = Bool(" << ValueExtract(*item.second).toBool() << ")\n";
            break;
        case ValueHolder::Type::String:
            std::cout << '"' << item.first << "\" = String(" << ValueExtract(*item.second).toString() << ")\n";
            break;
        case ValueHolder::Type::Map:
            std::cout << '"' << item.first << "\" = Map\n";
            break;
        }
    }
    std::cout << "--- End map dump ---" << std::endl;
    return {};
}

ValueHolder mapkeys(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("mapkeys", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::Map)
        throwNotAMap("mapkeys", 1);

    ValueArray& arr = ValueExtract(args[0]).toMap();
    ValueArray ret;
    ret.reserve(arr.size());
    int c = 0;
    for (const auto& [k,v] : arr) {
        ret.emplace(std::to_string(c), new ValueHolder(k));
        ++c;
    }
    ret.emplace("_list_end_", new ValueHolder(c));
    return ret;
}

}
