/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "StringUtils.h"
#include "Script/ValueExtract.h"
#include "Script/ScriptException.h"
#include "Error.h"
#include <regex>

using namespace Builtin::Error;

namespace  {
const std::regex rx_resemblesint(R"(^\d+$)");
const std::regex rx_resemblesreal(R"(^\d+\.\d+$)");
const std::regex rx_resemblesbool(R"(^(true|false|0|1)$)");
}

namespace Builtin::StringUtils {

ValueHolder strupper(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("strupper", args.size(), 1);

    const auto& S = ValueExtract(args[0]).toString();
    std::string ret;
    std::for_each(S.begin(), S.end(), [&ret](const char c){
        ret += std::toupper(c);
    });
    return ret;
}

ValueHolder strlower(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("strlower", args.size(), 1);

    const auto& S = ValueExtract(args[0]).toString();
    std::string ret;
    std::for_each(S.begin(), S.end(), [&ret](const char c){
        ret += std::tolower(c);
    });
    return ret;
}

ValueHolder strlen(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("strlen", args.size(), 1);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("strlen", 1);

    return static_cast<int>(ValueExtract(args[0]).toString().length());
}

ValueHolder strmid(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 3)
        throwInsufficientParameters("strmid", args.size(), 3);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("strmid", 1);

    if (args[1].getType() != ValueHolder::Type::Integer)
        throwNotAWholeNumber("strmid", 2);

    if (args[2].getType() != ValueHolder::Type::Integer)
        throwNotAWholeNumber("strmid", 3);

    const auto& S = ValueExtract(args[0]).toString();
    const auto start = ValueExtract(args[1]).toInt();
    const auto length = ValueExtract(args[2]).toInt();

    if (start < 0)
        throwNumberMustBePositive("strmid", 2);

    if (length < 0)
        throwNumberMustBePositive("strmid", 3);

    if (S.empty() || static_cast<unsigned>(start) > S.length() || length == 0)
        return std::string();

    return S.substr(static_cast<unsigned>(start), static_cast<unsigned>(length));
}

ValueHolder strsplit(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("strsplit", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("strsplit", 1);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("strsplit", 2);

    ValueArray ret;
    const auto& S = ValueExtract(args[0]).toString();
    const auto& delim = ValueExtract(args[1]).toString();

    if (delim.empty())
        throw ScriptException("strsplit: Delimiter is an empty string");

    const std::size_t delimlen = delim.size();
    std::size_t start = 0;
    std::size_t delimpos = 0;
    int idx = 0;
    while ((delimpos = S.find(delim, start)) != S.npos) {
        auto text = S.substr(start, delimpos - start);
        if (!text.empty())
            ret.emplace(std::to_string(idx++), new ValueHolder(text));
        start = delimpos + delimlen;
    }

    auto rest = S.substr(start);
    if (!rest.empty())
        ret.emplace(std::to_string(idx), new ValueHolder(rest));

    return std::move(ret);
}

ValueHolder strfind(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() < 2 || args.size() > 3)
        throwInsufficientParameters("strfind", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("strfind", 1);

    if (args[1].getType() != ValueHolder::Type::String)
        throwNotAString("strfind", 2);

    std::size_t start = 0;
    if (args.size() == 3) {
        if (args[2].getType() != ValueHolder::Type::Integer)
            throwNotAWholeNumber("strfind", 3);
        int val = ValueExtract(args[2]).toInt();
        if (val < 0)
            throwNumberMustBePositive("strfind", 3);
        start = static_cast<unsigned>(val);
    }

    const auto& S = ValueExtract(args[0]).toString();
    const auto& search = ValueExtract(args[1]).toString();
    if (search.empty())
        throw ScriptException("strfind: Searching for an empty string");
    auto pos = S.find(search, start);
    if (pos == S.npos)
        return -1;
    else
        return static_cast<int>(pos);
}

ValueHolder strleft(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("strleft", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("strleft", 1);

    if (args[1].getType() != ValueHolder::Type::Integer)
        throwNotAWholeNumber("strleft", 1);

    const auto& S = ValueExtract(args[0]).toString();
    const auto len = ValueExtract(args[1]).toInt();

    if (len < 0)
        throwNumberMustBePositive("strleft", 2);

    return S.substr(0, static_cast<unsigned>(len));
}

ValueHolder strright(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("strright", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("strright", 1);

    if (args[1].getType() != ValueHolder::Type::Integer)
        throwNotAWholeNumber("strright", 1);

    const auto& S = ValueExtract(args[0]).toString();
    const auto len = ValueExtract(args[1]).toInt();

    if (len < 0)
        throwNumberMustBePositive("strright", 2);

    int pos = static_cast<int>(S.length()) - len;
    if (pos < 0)
        pos = 0;

    return S.substr(static_cast<unsigned>(pos));
}

ValueHolder strrep(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 2)
        throwInsufficientParameters("strrep", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("strrep", 1);

    if (args[1].getType() != ValueHolder::Type::Integer)
        throwNotAWholeNumber("strrep", 2);

    const auto& S = ValueExtract(args[0]).toString();
    const auto count = ValueExtract(args[1]).toInt();

    if (count < 0)
        throwNumberMustBePositive("strright", 2);

    std::string ret;
    if (count == 0)
        return ret;
    for (int i = 0; i < count; ++i)
        ret += S;
    return ret;
}

ValueHolder resemblesint(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("resemblesint", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("resemblesint", 1);

    const auto& S = ValueExtract(args[0]).toString();
    if (S.empty())
        return false;

    std::smatch match;
    std::regex_match(S, match, rx_resemblesint);
    return match[0].length() != 0;
}

ValueHolder resemblesreal(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("resemblesreal", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("resemblesreal", 1);

    const auto& S = ValueExtract(args[0]).toString();
    if (S.empty())
        return false;

    std::smatch match;
    std::regex_match(S, match, rx_resemblesreal);
    return match[0].length() != 0;
}

ValueHolder resemblesbool(Script&, std::vector<ValueHolder>& args)
{
    if (args.size() != 1)
        throwInsufficientParameters("resemblesbool", args.size(), 2);

    if (args[0].getType() != ValueHolder::Type::String)
        throwNotAString("resemblesbool", 1);

    const auto& S = ValueExtract(args[0]).toString();
    if (S.empty())
        return false;

    std::smatch match;
    std::regex_match(S, match, rx_resemblesbool);
    return match[0].length() != 0;
}

}
