/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "Manager.h"
#include "ui_Manager.h"
#include "MdiManager.h"
#include "IWin/IWinStatus.h"
#include "ConfigMgr.h"
#include "ICommand/ICommand.h"
#include "ScriptFunctions/RegisterScriptFunctions.h"
#include <QDebug>
#include <QFile>
#include <QFileDialog>
#include <QMessageBox>

namespace {
std::optional<ScriptManager*> Instance;
}

ScriptManager::ScriptManager(MdiManager* mdiManager, QWidget *parent) :
    QDialog(parent),
    ui(new Ui::ScriptManager),
    mdiManager(mdiManager)
{
    ui->setupUi(this);
    registerScriptFunctions();
    ui->listView->setModel(&m_listModel);
}

void ScriptManager::commandHandler(const std::string& input)
{
    auto& connection = (*Instance)->m_context->getConnection();
    ICommand cmd(connection);
    cmd.parse(input.c_str());
}

void ScriptManager::event(const std::string& eventName, std::vector<ValueHolder>& params)
{
    for (Script& s : (*Instance)->m_scripts) {
        if (!s.dispatchEvent(eventName, params))
            printError(s.lastError());
    }
}

void ScriptManager::event(const std::string& eventName, QList<ValueHolder>& params)
{
    std::vector<ValueHolder> convp;
    for (auto& param : params)
        convp.emplace_back(param); // std::move?
    event(eventName, convp);
}

void ScriptManager::event(const std::string& eventName)
{
    std::vector<ValueHolder> emptyParams;
    event(eventName, emptyParams);
}

void ScriptManager::init(MdiManager* mdiManager, QWidget* parent)
{
    Instance.emplace(new ScriptManager(mdiManager, parent));
    ScriptManager::setContext(mdiManager->currentStatus());
    (*Instance)->loadAllScripts();
}

void ScriptManager::setContext(IWinStatus* context)
{
    if (!Instance.has_value()) {
        qWarning() << "Script manager was not initialized when tried to set context!";
        return;
    }
    instance()->m_context = context;
}

ScriptManager* ScriptManager::instance()
{
    return *Instance;
}

IWinStatus* ScriptManager::getContext()
{
    return (*Instance)->m_context;
}

ScriptManager::~ScriptManager()
{
    delete ui;
}

void ScriptManager::loadAllScripts()
{

    auto& cfg = ConfigMgr::instance();
    for (const QString& path : cfg.scripts()) {
        bool ok = loadFromFile(path);
        if (!ok) {
            auto btn = QMessageBox::warning(this, tr("Load script"), tr("The script located at:\n%1\nFailed to load. Keep the entry in the manager?").arg(path),
                                            QMessageBox::Yes | QMessageBox::No);
            if (btn == QMessageBox::No) {
                m_scripts.pop_back();
                continue;
            }
            else {
                m_listModel.addPath(path);
                m_listModel.markWarning(path);
            }
        }
        else {
            m_listModel.addPath(path);
        }
    }
}

void ScriptManager::contextMenuPopup(ScriptMenuType type, const QPoint& pos, SymbolScope& symbols)
{
    QString menuName;
    switch (type) {
        case ScriptMenuType::Status:
            menuName = "status";
            break;
        case ScriptMenuType::Channel:
            menuName = "channel";
            break;
        case ScriptMenuType::Memberlist:
            menuName = "memberlist";
            break;
        case ScriptMenuType::Private:
            menuName = "private";
            break;
    }
    if (!menuName.isEmpty()) {
        QMenu* menu = contextMenu(menuName, symbols);
        if (menu) menu->popup(pos);
    }
}

bool ScriptManager::runCommand(const QString& command)
{
    bool ret{ false };
    std::string cmd = command.toStdString();
    for (Script& script : m_scripts) {
        bool r = script.runCommand(cmd);
        if (script.hasError()) {
            printError(script.lastError());
            continue;
        }
        if (!ret)
            ret = r;
    }
    return ret;
}

bool ScriptManager::loadFromFile(const QString& path)
{
    auto fixedPath = path;
    MAKE_FIXED_PATH(path, fixedPath);

    QFile f(fixedPath);
    if (!f.open(QIODevice::ReadOnly)) {
        QMessageBox::warning(this, tr("Load script"), tr("Unable to open script file from\n%1").arg(fixedPath));
        return false;
    }
    QByteArray data(f.readAll());
    f.close();
    return loadFromData(data);
}

bool ScriptManager::loadFromData(const QByteArray& data)
{
    auto& script = m_scripts.emplace_back(data.toStdString(), &commandHandler,
                                            [](const std::string& eventName, std::vector<ValueHolder>& params)
                                                { event(eventName, params); });
    bool ret = script.reload();
    if (!ret)
        mdiManager->currentStatus()->print(PrintType::ProgramInfo, script.lastError().c_str());
    return ret;
}

void ScriptManager::unloadAllScripts()
{
    m_scripts.clear();
}

QMenu* ScriptManager::contextMenu(const QString& name, SymbolScope& symbols)
{
    if (m_menu)
        m_menu->deleteLater();
    m_menu = new QMenu();

    bool contents = false;
    for (Script& script : m_scripts) {
        auto maybeMenu = script.menu(name.toStdString());
        if (!maybeMenu) continue;
        contents = true;
        populateMenu(maybeMenu->get().content(), m_menu, script, symbols);
    }
    if (!contents) {
        m_menu->deleteLater();
        m_menu = nullptr;
    }

    return m_menu;
}

void ScriptManager::populateMenu(const MenuEntry& submenu, QMenu* menu, Script& script, SymbolScope& symbols)
{
    for (const MenuEntry& entry : submenu.entries()) {
        if (entry.type() == MenuEntry::Type::Item) {
            menu->addAction(entry.label().c_str(), [&script, &entry, &symbols]{
                auto ret = script.runScopeWithSymbols(entry.scope(), symbols);
                if (!ret)
                    printError(script.lastError());
            });
        }
        else if (entry.type() == MenuEntry::Type::Submenu) {
            QMenu* m = menu->addMenu(entry.label().c_str());
            populateMenu(entry, m, script, symbols);
        }
        else {
            menu->addSeparator();
        }
    }
}

void ScriptManager::printError(const std::string& msg)
{
    (*Instance)->mdiManager->currentStatus()->print(PrintType::ProgramInfo, tr("Script error: %1").arg(msg.c_str()));
}

void ScriptManager::reloadScript(Script& script, const QString& pathToScript)
{
    auto fixedPath = pathToScript;
    MAKE_FIXED_PATH(pathToScript, fixedPath);

    QFile f(fixedPath);
    if (!f.open(QIODevice::ReadOnly)) {
        QMessageBox::warning(this, tr("Reload script"), tr("Unable to open script file for reloading\n%1").arg(fixedPath));
        m_listModel.markWarning(pathToScript);
        return;
    }
    QByteArray data(f.readAll());
    f.close();

    if (script.reload(data.toStdString()))
        m_listModel.unmarkWarning(pathToScript);
    else {
        printError(script.lastError());
        m_listModel.markWarning(pathToScript);
    }
}

void ScriptManager::on_btnLoad_clicked()
{
    QString path = QFileDialog::getOpenFileName(this, tr("Open script..."), LOCAL_PATH, tr("All files (*)"));
    if (path.isEmpty())
        return;

    QModelIndex existingIdx = m_listModel.pathIndex(path);
    if (existingIdx.isValid()) {
        auto btn = QMessageBox::warning(this, tr("Load script"), tr("The script located at:\n%1\nIs already loaded. Reload script instead?").arg(path),
                                        QMessageBox::Yes | QMessageBox::No);
        if (btn == QMessageBox::Yes) {
            auto it = m_scripts.begin();
            for (int i = 0; i < existingIdx.row(); ++i)
                ++it;

            reloadScript(*it, path);
        }
        return;
    }

    bool ok = loadFromFile(path);
    if (!ok) {
        auto btn = QMessageBox::warning(this, tr("Load script"), tr("The script located at:\n%1\nFailed to load. Keep the entry in the manager?").arg(path),
                                        QMessageBox::Yes | QMessageBox::No);
        if (btn == QMessageBox::No) {
            m_scripts.pop_back();
            return;
        }
    }

    m_listModel.addPath(path);
    ConfigMgr::instance().addScript(path);
    if (!ok)
        m_listModel.markWarning(path);
}

void ScriptManager::on_btnReload_clicked()
{
    auto* sel = ui->listView->selectionModel();
    auto modelidx = sel->currentIndex();
    if (!modelidx.isValid())
        return;

    auto it = m_scripts.begin();
    for (int i = 0; i < modelidx.row(); ++i)
        ++it;

    QString path = modelidx.data().toString();
    reloadScript(*it, path);
}

void ScriptManager::on_btnRemove_clicked()
{
    auto* sel = ui->listView->selectionModel();
    auto modelidx = sel->currentIndex();
    if (!modelidx.isValid())
        return;

    {
        const auto button = QMessageBox::question(this, tr("Unload script"), tr("Are you sure you want to unload the selected script?"));
        if (button == QMessageBox::No)
            return;
    }

    auto it = m_scripts.begin();
    for (int i = 0; i < modelidx.row(); ++i)
        ++it;

    QVariant path = m_listModel.data(modelidx);
    ConfigMgr::instance().delScript(path.toString());
    m_listModel.delPath(path.toString());
    m_scripts.erase(it);
}
