/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#ifndef MANAGER_H
#define MANAGER_H

#include "Script.h"
#include "ICommand/ICommand.h"
#include "ManagerListModel.h"
#include <QDialog>
#include <QMenu>
#include <list>

using ScriptArray = std::unordered_map<std::string, ValueHolder>;

enum class ScriptMenuType {
    Status,
    Channel,
    Memberlist,
    Private,
    Menubar,
    Custom
};

namespace Ui {
class ScriptManager;
}

class MdiManager;
class ScriptManager : public QDialog
{
    Q_OBJECT

public:
    static void init(MdiManager* mdiManager, QWidget* parent);
    static void setContext(IWinStatus* context);
    static ScriptManager* instance();
    static IWinStatus* getContext();
    ~ScriptManager();
    void loadAllScripts();
    void contextMenuPopup(ScriptMenuType type, const QPoint& pos, SymbolScope& symbols);
    bool runCommand(const QString& command);
    bool loadFromFile(const QString& path);
    bool loadFromData(const QByteArray& data);
    void unloadAllScripts();

    static void event(const std::string& eventName, std::vector<ValueHolder>& params);
    static void event(const std::string& eventName, QList<ValueHolder>& params);
    static void event(const std::string& eventName);

    static void printError(const std::string& msg);

private slots:
    void on_btnLoad_clicked();
    void on_btnReload_clicked();
    void on_btnRemove_clicked();

private:
    explicit ScriptManager(MdiManager* mdiManager, QWidget *parent = nullptr);
    static void commandHandler(const std::string& input);
    QMenu* contextMenu(const QString& name, SymbolScope& symbols);
    void populateMenu(const MenuEntry& submenu, QMenu* menu, Script& script, SymbolScope& symbols);
    void reloadScript(Script& script, const QString& pathToScript);
    Ui::ScriptManager *ui;
    MdiManager* mdiManager;
    std::list<Script> m_scripts;
    QMenu* m_menu{ nullptr };
    IWinStatus* m_context;
    ManagerListModel m_listModel;
};

namespace {
[[maybe_unused]]
int convVal(int v) { return v; }

[[maybe_unused]]
double convVal(double v) { return v; }

[[maybe_unused]]
bool convVal(bool v) { return v; }

[[maybe_unused]]
std::string convVal(const std::string& v) { return v; }

[[maybe_unused]]
std::string convVal(const QString& v) { return v.toStdString(); }

[[maybe_unused]]
ValueArray convVal(const ValueArray& v)
{
    ValueArray copy;
    for (const auto& [key,val] : v)
        copy.emplace(key, new ValueHolder(*val.get()));
    return copy;
}

[[maybe_unused]]
ValueArray convVal(const ScriptArray& v)
{
    ValueArray copy;
    for (const auto& [key,val] : v)
        copy.emplace(key, new ValueHolder(val));
    return copy;
}
}

template<typename ... Ts>
void scriptEvent(const std::string& eventName, const Ts& ... args)
{
    if constexpr(sizeof...(args) > 0) {
        std::vector<ValueHolder> vals{ ValueHolder(convVal(args))... };
        ScriptManager::event(eventName, vals);
    }
    else {
        std::vector<ValueHolder> vals;
        ScriptManager::event(eventName, vals);
    }
}

template<typename ... Ts>
void contextualScriptEvent(IWinStatus* context, const std::string& eventName, const Ts& ... args)
{
    auto* prevContext = ScriptManager::getContext();
    ScriptManager::setContext(context);
    scriptEvent(eventName, args...);
    ScriptManager::setContext(prevContext);
}

#endif // MANAGER_H
