/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "ManagerListModel.h"
#include <QDebug>
#include <QIcon>

void ManagerListModel::addPath(const QString& path)
{
    if (m_list.contains(path))
        return;

    int idx = m_list.size();

    beginInsertRows(createIndex(idx, 0, this), idx, idx);
    m_list.push_back(path);
    endInsertRows();
}

void ManagerListModel::delPath(const QString& path)
{
    if (!m_list.contains(path))
        return;

    int idx = findEntry(path);
    if (idx < 0)
        return;

    beginRemoveRows(createIndex(idx, 0, this), idx, idx);
    m_warnings.removeAll(idx);
    m_list.removeAll(path);
    endRemoveRows();
}

QModelIndex ManagerListModel::pathIndex(const QString& path)
{
    int row = findEntry(path);
    if (row < 0)
        return QModelIndex();
    else
        return createIndex(row, 0, this);
}

void ManagerListModel::markWarning(const QString& entry)
{
    int idx = findEntry(entry);
    if (idx < 0) return;

    beginInsertRows(QModelIndex(), idx, idx);
    if (!isWarning(idx))
        m_warnings.push_back(idx);
    endInsertRows();
}

void ManagerListModel::unmarkWarning(const QString& entry)
{
    int idx = findEntry(entry);
    if (idx < 0) return;

    beginInsertRows(QModelIndex(), idx, idx);
    m_warnings.removeAll(idx);
    endInsertRows();
}

QVariant ManagerListModel::data(const QModelIndex& index, int role) const
{
    if (index.row() < 0 || index.row() >= m_list.size())
          return QVariant();

    if (role == Qt::DisplayRole) {
        return m_list[index.row()];
    }
    else if (role == Qt::SizeHintRole) {
        return QSize(0, 30);
    }
    else if (role == Qt::DecorationRole && isWarning(index.row())) {
        return QIcon(":/Icons/exclamation.png");
    }

    return QVariant();
}

int ManagerListModel::rowCount(const QModelIndex&) const
{
    return m_list.size();
}

int ManagerListModel::findEntry(const QString& entry) const
{
    QString entryFixed = entry;
    MAKE_FIXED_PATH(entry, entryFixed);

    bool found = false;
    int idx = 0;
    for (const QString& item : m_list) {
        QString itemFixed = item;
        MAKE_FIXED_PATH(item, itemFixed);

        if (itemFixed == entryFixed) {
            found = true;
            break;
        }
        ++idx;
    }

    if (!found)
        return -1;
    else
        return idx;
}

bool ManagerListModel::isWarning(int idx) const
{
    return m_warnings.contains(idx);
}
