/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "Menu.h"
#include "Tokens.h"

struct MenuPriv
{
    explicit MenuPriv(const std::string& name_);
    std::string name;
    MenuEntry content;
};

struct MenuEntryPriv
{
    MenuEntryPriv(const std::string& label_, MenuEntry::Type type_)
        : label(label_)
        , type(type_)
    {}

    virtual ~MenuEntryPriv() = default;
    const std::string label;
    const MenuEntry::Type type;
};

struct SubmenuEntryPriv : MenuEntryPriv
{
    explicit SubmenuEntryPriv(const std::string& label_)
        : MenuEntryPriv(label_, MenuEntry::Type::Submenu)
    {}

    ~SubmenuEntryPriv() override = default;

    std::vector<MenuEntry> entries;
};

struct DivideEntryPriv : MenuEntryPriv
{
    DivideEntryPriv()
        : MenuEntryPriv("", MenuEntry::Type::Divide)
    {}

    ~DivideEntryPriv() = default;
};

struct ItemEntryPriv : MenuEntryPriv
{
    ItemEntryPriv(const std::string& label_, Scope&& scope)
        : MenuEntryPriv(label_, MenuEntry::Type::Item)
        , scope(std::move(scope))
    {}

    ~ItemEntryPriv() = default;

    Scope scope;
};

Menu::~Menu() = default;

MenuPriv::MenuPriv(const std::string& name_)
    : name(name_)
    , content(name_)
{}

Menu::Menu(const std::string& name)
    : mp(new MenuPriv(name))
{}

Menu::Menu(Menu&& other)
    : mp(std::move(other.mp))
{}

Menu& Menu::operator=(Menu&& other)
{
    mp = std::move(other.mp);
    return *this;
}

const MenuEntry& Menu::content() const
{
    return mp->content;
}

MenuEntry& Menu::content()
{
    return mp->content;
}

const std::string Menu::name() const
{
    return mp->name;
}

MenuEntry::MenuEntry()
    : mp(new DivideEntryPriv())
{}

MenuEntry::MenuEntry(const std::string& label)
    : mp(new SubmenuEntryPriv(label))
{}

MenuEntry::MenuEntry(const std::string& label, Scope&& scope)
    : mp(new ItemEntryPriv(label, std::move(scope)))
{}

MenuEntry::MenuEntry(MenuEntry&& other)
    : mp(std::move(other.mp))
{}

MenuEntry::~MenuEntry() = default;

const Scope& MenuEntry::scope() const
{
    auto* imp = dynamic_cast<ItemEntryPriv*>(mp.get());
    return imp->scope;
}

const std::vector<MenuEntry>& MenuEntry::entries() const
{
    auto* imp = dynamic_cast<SubmenuEntryPriv*>(mp.get());
    return imp->entries;
}

const std::string& MenuEntry::label() const
{
    return mp->label;
}

MenuEntry& MenuEntry::addSubMenu(const std::string& label)
{
    auto* imp = dynamic_cast<SubmenuEntryPriv*>(mp.get());
    imp->entries.emplace_back(label);
    return imp->entries.back();
}

MenuEntry& MenuEntry::addItem(const std::string& label, Scope&& scope)
{
    auto* imp = dynamic_cast<SubmenuEntryPriv*>(mp.get());
    imp->entries.emplace_back(label, std::move(scope));
    return imp->entries.back();
}

MenuEntry& MenuEntry::addDivide()
{
    auto* imp = dynamic_cast<SubmenuEntryPriv*>(mp.get());
    imp->entries.emplace_back();
    return imp->entries.back();
}

MenuEntry::Type MenuEntry::type() const
{
    return mp->type;
}

Scope& MenuEntry::scope()
{
    auto* imp = dynamic_cast<ItemEntryPriv*>(mp.get());
    return imp->scope;
}

std::vector<MenuEntry>& MenuEntry::entries()
{
    auto* imp = dynamic_cast<SubmenuEntryPriv*>(mp.get());
    return imp->entries;
}
