/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "ParserToken.h"
#include <unordered_map>

namespace {
using Op = ParserOperator;
std::unordered_map<ParserOperator, int> PrecedenceMap {
    { Op::Eq, 1 },
    { Op::Ne, 1 },

    { Op::Lt, 2 },
    { Op::Gt, 2 },
    { Op::Lte, 2 },
    { Op::Gte, 2 },

    { Op::Multi, 3 },
    { Op::Div, 3 },
    { Op::Mod, 3 },

    { Op::Plus, 4 },
    { Op::Minus, 4 },

    { Op::LAnd, 5 },
    { Op::LOr, 6 },

    { Op::Assign, 7 }
};
}

ParserToken::ParserToken(ParserToken&& other)
    : m_order(other.m_order)
    , m_tk(std::move(other.m_tk))
    , m_tokenbase(other.m_tokenbase)
{}

ParserToken::ParserToken(const ParserToken& other)
    : m_order(other.m_order)
    , m_tk(other.m_tk)
    , m_tokenbase(other.m_tokenbase)
{}

ParserToken::ParserToken(const TokenBase& tokenbase, ParserToken::Operator op)
    : m_tk(op)
    , m_tokenbase(tokenbase)
{}

ParserToken::ParserToken(const TokenBase& tokenbase, const ValueHolder& val)
    : m_tk(val)
    , m_tokenbase(tokenbase)
{}

ParserToken& ParserToken::operator=(const ParserToken& other)
{
     m_tk = other.m_tk;
     m_order = other.m_order;
     return *this;
}

ParserToken::Type ParserToken::type() const
{
    switch (m_tk.index()) {
    case 0: return Type::Value;
    default: return Type::Operator;
    }
}

ValueHolder& ParserToken::value()
{
    return std::get<ValueHolder>(m_tk);
}

const ValueHolder& ParserToken::value() const
{
    return std::get<ValueHolder>(m_tk);
}

ParserToken::Operator ParserToken::oper() const
{
    return std::get<Operator>(m_tk);
}

ParserToken::Operator ParserToken::strToOperator(const std::string& oper)
{
    if (oper == "=")
        return Operator::Assign;

    else if (oper == "<")
        return Operator::Lt;

    else if (oper == ">")
        return Operator::Gt;

    else if (oper == "<=")
        return Operator::Lte;

    else if (oper == ">=")
        return Operator::Gte;

    else if (oper == "==")
        return Operator::Eq;

    else if (oper == "!=")
        return Operator::Ne;

    else if (oper == "+")
        return Operator::Plus;

    else if (oper == "-")
        return Operator::Minus;

    else if (oper == "*")
        return Operator::Multi;

    else if (oper == "/")
        return Operator::Div;

    else if (oper == "%")
        return Operator::Mod;

    else if (oper == "&&")
        return Operator::LAnd;

    else if (oper == "||")
        return Operator::LOr;

    else
        return Operator::Invalid;
}

int ParserToken::operPrecedence(Operator oper)
{
    return PrecedenceMap.at(oper);
}

void ParserToken::setOrder(int order)
{
    m_order = order;
}

int ParserToken::order() const
{
    return m_order;
}

const TokenBase &ParserToken::tokenBase() const
{
    return m_tokenbase;
}
