/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "ParserToken.h"
#include "Script.h"
#include "ScriptException.h"
#include "Builtin/Builtin.h"
#include "ValueExtract.h"
#include "Script/Builtin/ListUtils.h" // Needed for LIST_END_MAGIC
#include <iostream>
#include <sstream>

namespace {
ValueHolder calculate(const ValueHolder& lhs, ParserOperator oper, const ValueHolder& rhs)
{
    using op = ParserOperator;
    switch (oper) {
    /*
      case op::Assign:
        return lhs = rhs;
    */
    case op::Lt:
        return lhs < rhs;
    case op::Gt:
        return lhs > rhs;
    case op::Lte:
        return lhs <= rhs;
    case op::Gte:
        return lhs >= rhs;
    case op::Eq:
        return lhs == rhs;
    case op::Ne:
        return lhs != rhs;
    case op::Plus:
        return lhs + rhs;
    case op::Minus:
        return lhs - rhs;
    case op::Multi:
        return lhs * rhs;
    case op::Div:
        return lhs / rhs;
    case op::Mod:
        return lhs % rhs;
    case op::LAnd:
        return lhs && rhs;
    case op::LOr:
        return lhs || rhs;
    default:
        return {};
    }
}
}

namespace EMSG {
constexpr auto* UndefinedError = "Undefined error"; // Use sparingly, and only when a case seems illogical and unlikely to happen.
constexpr auto* IdentNotFound = "Identifier not found: ";
constexpr auto* NotAFunction = "Identifier not callable: ";
constexpr auto* InvalidParamCount = "Invalid parameter count";
constexpr auto* CannotAssign = "Cannot assign with that operator";
constexpr auto* ExprOperStart = "Expression starts with operator";
constexpr auto* ExprOperEnd = "Expression ends with operator";
constexpr auto* NotAValue = "Not a value: ";
constexpr auto* SyntaxError = "Syntax error";
constexpr auto* ExpectedOperator = "Expected an operator";
constexpr auto* NotAnOperator = "Not an operator: ";
constexpr auto* ElseWithoutPrevIf = "'else' without a previous 'if'";
constexpr auto* ContinueOutsideWhile = "'continue' outside a 'while'";
constexpr auto* BreakOutsideWhile = "'break' outside a 'while'";
constexpr auto* SubscriptIsVoid = "Index/subscript of map resolved to void";
constexpr auto* IdentifierNotMap = "Identifier is not a map: ";
constexpr auto* ExpectedOneIdentifierForReference = "Expected only one identifier when assigning a reference";
constexpr auto* ExpectedOneIdentifierForList = "Expected only one identifier when assigning a list";
constexpr auto* CannotMixReferences = "Cannot mix a 'reference-to' operation in an expression";
}

SymbolScope Script::m_universal;

Script::Script(const std::string &input, CommandHandler commandHandler_, DispatchHandler dispatchHandler_)
    : m_data(input)
    , commandHandler(commandHandler_)
    , dispatchHandler(dispatchHandler_)
{
    registerBuiltinFunctions();
}

Script::Script(Script&& other)
    : m_glob(std::move(other.m_glob))
    , m_cmds(std::move(other.m_cmds))
    , m_hooks(std::move(other.m_hooks))
    , m_dlgs(std::move(other.m_dlgs))
    , m_dlgHndlCount(other.m_dlgHndlCount)
    , m_tokenizer(std::move(other.m_tokenizer))
    , m_data(std::move(other.m_data))
    , commandHandler(std::move(other.commandHandler))
    , dispatchHandler(std::move(other.dispatchHandler))
    , m_lastError(std::move(other.m_lastError))
    , m_lastErrorLinecol(other.m_lastErrorLinecol)
{}

Script::~Script()
{
    for (auto& [dlgName, dlg] : m_dlgs) {
        dlg.close();
    }
}

bool Script::reload(const std::string& newInput)
{
    m_data = newInput;
    return reload();
}

bool Script::reload()
{
    if (m_tokenizer) {
        m_dlgs.clear();
    	m_hooks.reset();
        m_cmds.reset();
        m_glob.reset();
        m_tokenizer.reset();
    }

    m_tokenizer.emplace(m_data);
    m_glob.emplace(&m_universal);
    m_cmds.emplace();
    m_hooks.emplace();

    resetError();

    if (!m_tokenizer->process()) {
        m_lastErrorLinecol = m_tokenizer->currentLineCol();
        m_lastError = m_tokenizer->lastError();
        return false;
    }

    // Load all global symbols to m_glob
    for (Scope& scope : *m_tokenizer) {
        try {
            auto& imps = dynamic_cast<ImperativeScope&>(scope);
            switch (imps.type()) {
            	case ImperativeBlockType::Function:
                    m_glob->set(imps.identifier(), scope);
                    break;
                case ImperativeBlockType::Command:
                    m_cmds->set(imps.identifier(), scope);
                    break;
                case ImperativeBlockType::Hook:
                    m_hooks->set(imps.identifier(), scope);
                    break;
            }
        } catch (std::bad_cast&) {} // Conversion error is not an error.
    }

    /* Load globals */
    {
        ValueHolder VoidVal;
        for (const std::string& identifier : m_tokenizer->globals())
            m_glob->set(identifier, VoidVal);
    }

    m_glob->set("void", ValueHolder());

    std::vector<ValueHolder> emptyParams;
    return dispatchEvent("load", emptyParams);
}

std::optional<ValueHolder> Script::runFunction(const std::string &identifier, std::vector<ValueHolder>& paramValues)
{
    SymbolScope symbols(&(*m_glob));
    resetError();

    auto symbolMaybe = symbols.get(identifier);
    if (!symbolMaybe) {
        m_lastError = EMSG::IdentNotFound + identifier;
        return std::nullopt;
    }

    auto& symbol = symbolMaybe->get();

    if (symbol.getType() == Symbol::Type::Builtin) {
        try {
            return symbol.runBuiltin(*this, paramValues);
        }  catch (const ScriptException& e) {
            m_lastErrorLinecol = {0, 0};
            m_lastError = e.m_originalMessage;
            return std::nullopt;
        }
    }

    if (symbol.getType() != Symbol::Type::Scope) {
        m_lastError = EMSG::NotAFunction;
        return std::nullopt;
    }

    try { return resolveParamsRunScope(symbol.getScope(), paramValues, symbols); }
    catch (const ScriptException& ex) {
        m_lastError = ex.m_originalMessage;
        m_lastErrorLinecol = ex.m_linecol;
    }
    return std::nullopt;
}

bool Script::runCommand(const std::string& input)
{
    resetError();
    auto maybe = m_cmds->get(input.substr(0, input.find_first_of(' ')));
    if (!maybe)
        return false;
    else {
        SymbolScope symbols(&(*m_glob));
        auto& symbol = maybe->get();

        const ImperativeScope& impScope = static_cast<const ImperativeScope&>(symbol.getScope());

        std::vector<ValueHolder> paramValues;
        bool fillLastParam = false;
        const std::size_t expectedParamCount = impScope.params().countStatements();
        std::string buf;
        std::size_t bufpos = input.find_first_of(' ');
        for (; bufpos < input.size(); ++bufpos) {
            const char c = input[bufpos];
            if (buf.empty() && isspace(c))
                continue;

            if (paramValues.size() == expectedParamCount - 1) {
                fillLastParam = true;
                break;
            }

            if (!buf.empty() && (isspace(c) || bufpos == input.size() - 1)) {
                if (bufpos == input.size() - 1)
                    buf += c;

                if (isdigit(buf[0])) {
                    if (buf.find_first_of('.') != std::string::npos)
                        paramValues.emplace_back( std::stod(buf) );
                    else
                        paramValues.emplace_back( std::stoi(buf) );
                }
                else
                    paramValues.emplace_back(buf);

                buf.clear();
            }
            else
                buf += c;
        }

        if (fillLastParam)
            paramValues.emplace_back(input.substr(bufpos));

        while (paramValues.size() < expectedParamCount)
            paramValues.emplace_back();

        try { resolveParamsRunScope(symbol.getScope(), paramValues, symbols); }
        catch (const ScriptException& ex) {
            m_lastError = ex.m_originalMessage;
            m_lastErrorLinecol = ex.m_linecol;
        }

        return true;
    }
}

bool Script::dispatchEvent(const std::string& event, std::vector<ValueHolder>& paramValues)
{
    resetError();
    auto symbolMaybe = m_hooks->get(event);
    if (!symbolMaybe) return true; // Hook not defined is not an error.

    auto& symbol = symbolMaybe->get();
    if (symbol.getType() != Symbol::Type::Scope) {
        m_lastError = EMSG::UndefinedError;
        m_lastErrorLinecol = { 0, 0 };
        return false;
    }

    SymbolScope symbols(&(m_glob.value()));
    try { resolveParamsRunScope(symbol.getScope(), paramValues, symbols); }
    catch (const ScriptException& ex) {
        m_lastError = ex.m_originalMessage;
        m_lastErrorLinecol = ex.m_linecol;
        return false;
    }
    return true;
}

std::optional<ValueHolder> Script::runScopeWithSymbols(const Scope& scope, SymbolScope& symbols)
{
    resetError();
    symbols.setParent(*m_glob);
    try {
        auto runresult = runScope(scope, symbols);
        auto ret = std::get<std::optional<ValueHolder>>(runresult);
        if (ret)
            return *ret;
        else
            return ValueHolder();
    }
    catch (const ScriptException& ex) {
        m_lastError = ex.m_originalMessage;
        m_lastErrorLinecol = ex.m_linecol;
        return std::nullopt;
    }
}

std::string Script::lastError() const
{
    std::stringstream errstr;
    if (m_lastErrorLinecol.first != 0)
        errstr << "[" << m_lastErrorLinecol.first << ":" << m_lastErrorLinecol.second << "] ";
    errstr << m_lastError;
    return errstr.str();
}

bool Script::hasError() const
{
    return !m_lastError.empty();
}

const std::optional<ConstMenuRef> Script::menu(const std::string& name) const
{
    const auto& menus = m_tokenizer->menus();
    auto iter = std::find_if(menus.begin(), menus.end(), [&name](const Menu& m){
        return m.name() == name;
    });

    if (iter != menus.end())
        return *iter;
    else
        return std::nullopt;
}

void Script::registerUniversalValue(const std::string& name, ValueHolder&& value)
{
    m_universal.set(name, std::move(value));
}

void Script::registerUniversalFunction(const std::string& name, const BuiltinFn& fnptr)
{
    m_universal.set(name, fnptr);
}

int Script::dialogConstruct(const std::string& name)
{
    const auto& dialogs = m_tokenizer->dialogs();
    auto iter = std::find_if(dialogs.begin(), dialogs.end(), [&name](const Dialog& d){
        return d.name() == name;
    });

    if (iter == dialogs.end())
        throw ScriptException("dialog: Dialog with name " + name + " is not defined");

    const Dialog& dialogTemplate = *iter;

    int handle = ++m_dlgHndlCount;
    m_dlgs.emplace(handle, ScriptDialog{ dialogTemplate, *this, handle });
    return handle;
}

ValueHolder Script::dialogIsValidHandle(int handle)
{
    return m_dlgs.count(handle) != 0;
}

void Script::dialogClose(int handle)
{
    try {
        ScriptDialog& dialog = m_dlgs.at(handle);
        dialog.close();
        m_dlgs.erase(handle);
    }  catch (std::out_of_range&) {
        throw ScriptException("dlgclose: Invalid dialog handle");
    }
}

void Script::dialogHide(int handle)
{
    try {
        ScriptDialog& dialog = m_dlgs.at(handle);
        dialog.hide();
    }  catch (std::out_of_range&) {
        throw ScriptException("dlgclose: Invalid dialog handle");
    }
}

void Script::dialogShow(int handle)
{
    try {
        ScriptDialog& dialog = m_dlgs.at(handle);
        dialog.show();
    }  catch (std::out_of_range&) {
        throw ScriptException("dlgclose: Invalid dialog handle");
    }
}

ValueArray Script::dialogList(const std::string& name)
{
    ValueArray ret;
    int idx = 0;
    for (auto& [hndl, dlg] : m_dlgs) {
        if (dlg.name() == name) {
            ret.emplace(std::to_string(idx), new ValueHolder(hndl));
            ++idx;
        }
    }
    ret.emplace(LIST_END_MAGIC, new ValueHolder(idx));
    return ret;
}

void Script::dialogSetAttr(int handle, const std::string& attribute, ValueHolder& value)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        dlg.setAttr(attribute, value);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

ValueHolder Script::dialogGetAttr(int handle, const std::string& attribute)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        return dlg.attr(attribute);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

void Script::dialogSetWidgetAttr(int handle, const std::string& widgetName, const std::string& attribute, ValueHolder& value)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        dlg.setWidgetAttr(widgetName, attribute, value);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

ValueHolder Script::dialogGetWidgetAttr(int handle, const std::string& widgetName, const std::string& attribute)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        return dlg.widgetAttr(widgetName, attribute);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

int Script::dialogTableWidgetCount(int handle, const std::string& widgetName)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        return dlg.tableWidgetCount(widgetName);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

void Script::dialogTableWidgetInsert(int handle, const std::string& widgetName, const ValueArray& cols)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        return dlg.tableWidgetInsert(widgetName, cols);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

void Script::dialogTableWidgetRemove(int handle, const std::string& widgetName, int index)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        return dlg.tableWidgetRemove(widgetName, index);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

int Script::dialogTableWidgetSelected(int handle, const std::string& widgetName)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        return dlg.tableWidgetSelected(widgetName);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

ValueArray Script::dialogTableWidgetRow(int handle, const std::string& widgetName, int index)
{
    try {
        ScriptDialog& dlg = m_dlgs.at(handle);
        return dlg.tableWidgetRow(widgetName, index);
    }  catch (const std::out_of_range&) {
        throw ScriptException("Invalid dialog handle");
    }
}

std::variant<std::optional<ValueHolder>, TokenBaseRef> Script::runScope(const Scope& scope, SymbolScope& parentSymbols)
{
    SymbolScope symbols(&parentSymbols);
    int lastConditional = -1; // -1 = No 'if' preceeds, 0 = false expression, 1 = true expression

    for (Expression& expr : scope) {
        if (expr.count() == 0) {
            std::cout << "FIXME: expr.count() == 0" << std::endl;
            continue;
        }

        /*
         * Commands
        */
        else if (expr[0].type() == STokenType::Command) {
            lastConditional = -1;
            const Token& cmd = static_cast<Token&>(expr[0]);
            std::string input = resolveCommandLine(cmd.token(), symbols);
            commandHandler(input);
        }

        /*
         * Assignment
        */
        else if (expr.count() == 3 && expr[0].type() == STokenType::Keyword && expr[1].type() == STokenType::Operator) {
            lastConditional = -1;

            const Token& lv = static_cast<Token&>(expr[0]);
            const Token& operToken = static_cast<Token&>(expr[1]);
            Expression& subex = static_cast<Expression&>(expr[2]);

            auto pto = ParserToken::strToOperator(operToken.token());

            if (expr[2].type() == STokenType::Reference) {
                if (pto != ParserOperator::Assign)
                    throw ScriptException(operToken.lineCol(), EMSG::CannotAssign);
                if (subex.count() != 1)
                    throw ScriptException(operToken.lineCol(), EMSG::ExpectedOneIdentifierForReference);

                const Token& referToToken = static_cast<Token&>(*subex.begin());
                const auto& referToStr = referToToken.token();

                auto referSym = symbols.get(referToStr);
                if (!referSym)
                    throw ScriptException(referToToken.lineCol(), EMSG::IdentNotFound + referToStr);

                ValueRef valRef = referSym->get().getValue();
                symbols.set(lv.token(), ValueHolder(valRef));
            }
            if (expr[2].type() == STokenType::MapKeyword) {
                if (pto != ParserOperator::Assign)
                    throw ScriptException(operToken.lineCol(), EMSG::CannotAssign);
                if (subex.count() != 1)
                    throw ScriptException(operToken.lineCol(), EMSG::ExpectedOneIdentifierForList);
            }
            else {
                if (pto == ParserOperator::Assign) {
                    auto maybeSymbol = symbols.get(lv.token());
                    if (maybeSymbol)
                        maybeSymbol->get().getValue() = resolveExpression(subex, symbols);
                    else
                        symbols.set(lv.token(), resolveExpression(subex, symbols));
                }
                else
                    throw ScriptException(operToken.lineCol(), EMSG::CannotAssign);
            }
        }

        /*
         * Assignment, map entry
        */
        else if (expr.count() == 3 && expr[0].type() == STokenType::MapKeyword && expr[1].type() == STokenType::Operator) {
            lastConditional = -1;
            const MapToken& lv = static_cast<MapToken&>(expr[0]);
            const Token& operToken = static_cast<Token&>(expr[1]);
            auto pto = ParserToken::strToOperator(operToken.token());
            if (pto == ParserOperator::Assign) {
                Expression& subex = static_cast<Expression&>(expr[2]);
                ValueHolder subscript = resolveExpression(lv.subscript(), symbols);
                ValueHolder value = resolveExpression(subex, symbols);
                if (subscript.getType() == ValueHolder::Type::Void)
                    throw ScriptException(lv.subscript().lineCol(), EMSG::SubscriptIsVoid);
                auto arrayMaybe = symbols.get(lv.token());

                if (arrayMaybe)
                    arrayMaybe->get().getValue().setArray(subscript.toStdString(), value);
                else {
                    symbols.set(subscript.toStdString(), ValueArray{});
                    arrayMaybe = symbols.get(lv.token());
                    if (arrayMaybe)
                        arrayMaybe->get().getValue().setArray(subscript.toStdString(), value);
                    else
                        throw ScriptException(lv.lineCol(), EMSG::IdentNotFound + lv.token());
                }
            }
            else
                throw ScriptException(operToken.lineCol(), EMSG::CannotAssign);
        }

        /*
         * Keywords
        */
        else if (expr[0].type() == STokenType::Keyword) {
            lastConditional = -1;
            const std::string& token = static_cast<Token&>(expr[0]).token();
            if (token == "return") { // TODO  own token type
                if (expr.count() == 1)
                    return std::make_optional<ValueHolder>();
                auto& retexpr = static_cast<Expression&>(expr[1]);
                return resolveExpression(retexpr, symbols);
            }
        }

        /*
         * Conditional: 'if'
        */
        else if (expr[0].type() == STokenType::CondIf) {
            ConditionalScopeToken& scopeTk = static_cast<ConditionalScopeToken&>(expr[0]);
            const Expression& condition = scopeTk.scope().condition();
            ValueHolder result = resolveExpression(condition, symbols);
            lastConditional = result != 0;
            if (lastConditional) {
                auto scopeResult = runScope(scopeTk.scope(), symbols);

                if (scopeResult.index() == 1)
                    return std::get<TokenBaseRef>(scopeResult).get();
                else {
                    auto& maybeResult = std::get<std::optional<ValueHolder>>(scopeResult);
                    if (maybeResult)
                        return *maybeResult;
                }
            }
        }

        /*
         * Conditional: 'else'
        */
        else if (expr[0].type() == STokenType::CondElse && lastConditional != 1) {
            if (lastConditional == -1)
                throw ScriptException(expr[0].lineCol(), EMSG::ElseWithoutPrevIf);

            ConditionalScopeToken& scopeTk = static_cast<ConditionalScopeToken&>(expr[0]);

            if (scopeTk.scope().begin()->begin()->type() == STokenType::CondIf) {
                ConditionalScopeToken& ifScopeTk = static_cast<ConditionalScopeToken&>( *(scopeTk.scope().begin()->begin()) );
                const Expression& condition = ifScopeTk.scope().condition();
                ValueHolder result = resolveExpression(condition, symbols);
                lastConditional = result != 0;
                if (lastConditional) {
                    auto scopeResult = runScope(ifScopeTk.scope(), symbols);

                    if (scopeResult.index() == 1)
                        return std::get<TokenBaseRef>(scopeResult).get();
                    else {
                        auto& maybeResult = std::get<std::optional<ValueHolder>>(scopeResult);
                        if (maybeResult)
                            return *maybeResult;
                    }
                }
            }
            else {
                lastConditional = -1;
                auto scopeResult = runScope(scopeTk.scope(), symbols);

                if (scopeResult.index() == 1)
                    return std::get<TokenBaseRef>(scopeResult).get();
                else {
                    auto& maybeResult = std::get<std::optional<ValueHolder>>(scopeResult);
                    if (maybeResult)
                        return *maybeResult;
                }
            }
        }

        /*
         * Conditional: 'while'
        */
        else if (expr[0].type() == STokenType::CondWhile) {
            ConditionalScopeToken& scopeTk = static_cast<ConditionalScopeToken&>(expr[0]);
            const Expression& condition = scopeTk.scope().condition();
            ValueHolder test = resolveExpression(condition, symbols);

            while (test != 0) {
                auto scopeResult = runScope(scopeTk.scope(), symbols);

                if (scopeResult.index() == 1) {
                    auto& control = std::get<TokenBaseRef>(scopeResult).get();
                    if (control.type() == STokenType::Continue)
                        continue;
                    else
                        break;
                }

                else {
                    auto& maybeResult = std::get<std::optional<ValueHolder>>(scopeResult);
                    if (maybeResult)
                        return *maybeResult;
                }

                test = resolveExpression(condition, symbols);
            }
        }

        /*
         * Control: 'continue', 'break'
        */
        else if (expr[0].type() == STokenType::Continue || expr[0].type() == STokenType::Break) {
            return expr[0];
        }

        /*
         * Function call where the result is discarded
        */
        else if (expr[0].type() == STokenType::Function) {
            resolveExpression(expr, symbols);
        }

        /*
         * Event dispatching
        */
        else if (expr[0].type() == STokenType::Dispatch) {
            auto& disptk = static_cast<DispatchToken&>(expr[0]);
            std::vector<ValueHolder> params = resolveParameters(disptk.parameters(), symbols);
            dispatchHandler(disptk.token(), params);
        }
    }

    return std::nullopt;
}

ValueHolder Script::resolveParamsRunScope(const Scope& scope, std::vector<ValueHolder>& paramValues, SymbolScope& symbols)
{
    const ImperativeScope& impScope = static_cast<const ImperativeScope&>(scope);
    std::size_t paramidx = 0;
    for (auto& paramExpr : impScope.params()) {
        if (impScope.params().countStatements() != paramValues.size()) {
            throw ScriptException(scope.getLineCol(), EMSG::InvalidParamCount);
        }
        else if (paramValues.size() > 0) {
            // TODO  implement default parameters here.
            auto& tokenbase = *(paramExpr.begin());
            if (tokenbase.type() == STokenType::Keyword) {
                const Token& token = static_cast<Token&>(tokenbase);
                symbols.set(token.token(), paramValues[paramidx]);
                ++paramidx;
            }
        }
    }

    auto scopeResult = runScope(scope, symbols);
    if (scopeResult.index() == 1) {
        auto& errTok = std::get<TokenBaseRef>(scopeResult).get();
        if (errTok.type() == STokenType::Continue)
            throw ScriptException(errTok.lineCol(), EMSG::ContinueOutsideWhile);
        else if (errTok.type() == STokenType::Break)
            throw ScriptException(errTok.lineCol(), EMSG::BreakOutsideWhile);
        else
            throw ScriptException(errTok.lineCol(), EMSG::UndefinedError);
    }
    else {
        auto& maybeResult = std::get<std::optional<ValueHolder>>(scopeResult);
        if (maybeResult)
            return *maybeResult;
        else
            return ValueHolder();
    }
}

ValueHolder Script::resolveExpression(const Expression& expr, SymbolScope& symbols)
{
    std::vector<ParserToken> parts;
    parts.reserve(expr.count() + 1);

    if (expr.begin() == expr.end())
        return {}; // Nothing to solve, return 'void'
    else {
        if (expr.begin()->type() == STokenType::Operator)
            throw ScriptException(expr.begin()->lineCol(), EMSG::ExprOperStart);

        if ((expr.end()-1)->type() == STokenType::Operator)
            throw ScriptException((expr.end()-1)->lineCol(), EMSG::ExprOperEnd);
    }

    if (expr.begin()->type() == STokenType::Reference) {
        if (expr.count() > 1)
            throw ScriptException((expr.end()-1)->lineCol(), EMSG::CannotMixReferences);

        try {
            auto& refToken = dynamic_cast<MapToken&>(*expr.begin());
            auto symbolMaybe = symbols.get(refToken.token());
            if (symbolMaybe) {
                ValueHolder subscriptVal = resolveExpression(refToken.subscript(), symbols);
                return ValueHolder(ValueRef(symbolMaybe->get().getValue()[subscriptVal.toStdString()]));
            }
            else
                return {};
        }  catch (const std::bad_cast&) {
            auto& refToken = static_cast<Token&>(*expr.begin());
            auto symbolMaybe = symbols.get(refToken.token());
            if (symbolMaybe)
                return ValueHolder(ValueRef(symbolMaybe->get().getValue()));
            else
                return {};
        }
    }

    /*
     * Add values to the parts list.
     * This is done recursively when hitting a subexpression.
    */
    for (TokenBase& tkbase : expr) {
        if (tkbase.type() == STokenType::Keyword) {
            const Token& tk = static_cast<const Token&>(tkbase);
            auto optval = symbols.get(tk.token());
            if (!optval)
                throw ScriptException(tkbase.lineCol(), EMSG::IdentNotFound + tk.token());
            else {
                Symbol& sym = *optval;
                if (sym.getType() != Symbol::Type::Value) {
                    throw ScriptException(tkbase.lineCol(), EMSG::NotAValue + tk.token());
                }
                else {
                    parts.emplace_back(tkbase, sym.getValue());
                }
            }
        }

        else if (tkbase.type() == STokenType::MapKeyword) {
            const MapToken& tk = static_cast<const MapToken&>(tkbase);
            auto optval = symbols.get(tk.token());
            if (!optval)
                throw ScriptException(tkbase.lineCol(), EMSG::IdentNotFound + tk.token());
            else {
                Symbol& sym = *optval;
                ValueHolder& array = sym.getValue();

                if (array.getType() != ValueHolder::Type::Map && array.getType() != ValueHolder::Type::String) {
                    throw ScriptException(tkbase.lineCol(), EMSG::IdentifierNotMap + tk.token());
                }
                else {
                    ValueHolder subscriptVal = resolveExpression(tk.subscript(), symbols);
                    if (subscriptVal.getType() == ValueHolder::Type::Void)
                        throw ScriptException(tk.subscript().lineCol(), EMSG::SubscriptIsVoid);
                    std::string subscript = subscriptVal.toStdString();
                    if (array.getType() == ValueHolder::Type::String) {
                        auto idx = std::stoi(subscript);
                        const std::string& value = ValueExtract(array).toString();
                        if (idx < 0 || idx > static_cast<int>(value.length()) - 1)
                            parts.emplace_back(tkbase, std::string());
                        else {
                            std::string character;
                            character.push_back(value[static_cast<unsigned>(idx)]);
                            parts.emplace_back(tkbase, character);
                        }
                    }
                    else {
                        parts.emplace_back(tkbase, array[subscript]);
                    }
                }
            }
        }

        else if (tkbase.type() == STokenType::NumberLiteral) {
            const Token& tk = static_cast<const Token&>(tkbase);
            parts.emplace_back(tkbase, ValueHolder{ std::stoi(tk.token()) });
        }
        else if (tkbase.type() == STokenType::RealLiteral) {
            const Token& tk = static_cast<const Token&>(tkbase);
            parts.emplace_back(tkbase, ValueHolder{ std::stod(tk.token()) });
        }
        else if (tkbase.type() == STokenType::StringLiteral) {
            const Token& tk = static_cast<const Token&>(tkbase);
            parts.emplace_back(tkbase, ValueHolder{ tk.token() });
        }
        else if (tkbase.type() == STokenType::BoolLiteral) {
            const BoolToken& tk = static_cast<const BoolToken&>(tkbase);
            parts.emplace_back(tkbase, ValueHolder{ tk.value() });
        }
        else if (tkbase.type() == STokenType::Operator) {
            const Token& tk = static_cast<const Token&>(tkbase);
            auto op = ParserToken::strToOperator(tk.token());
            if (op == ParserOperator::Invalid)
                throw ScriptException(tk.lineCol(), EMSG::NotAnOperator + tk.token());
            parts.emplace_back(tkbase, op);
        }
        else if (tkbase.type() == STokenType::Expression) {
            const Expression& subexpr = static_cast<const Expression&>(tkbase);
            parts.push_back(ParserToken(tkbase, resolveExpression(subexpr, symbols)));
        }
        else if (tkbase.type() == STokenType::Function) {
            try {
                const FunctionToken& funtk = dynamic_cast<const FunctionToken&>(tkbase);
                std::vector<ValueHolder> params = resolveParameters(funtk.parameters(), symbols);
                auto result = runFunction(funtk.identifier(), params);
                if (!result)
                    throw ScriptException(m_lastErrorLinecol, m_lastError);
                parts.emplace_back(tkbase, std::move(*result));
            }  catch (std::bad_cast&) {
                try {
                    const Token& tk = dynamic_cast<const Token&>(tkbase);
                    m_lastErrorLinecol = tkbase.lineCol();
                    throw ScriptException(m_lastErrorLinecol, EMSG::NotAFunction + tk.token());
                }  catch (std::bad_cast&) {
                    throw ScriptException(m_lastErrorLinecol, EMSG::NotAFunction + std::string("<unable to read token>"));
                }
            }
        }
        else {
            throw ScriptException(tkbase.lineCol(), EMSG::SyntaxError);
        }
    }


    /*
     * Find the operation order and set it
    */
    bool processing = true;
    int order = 0;

    while (processing) {
        processing = false;
        ParserToken* lpn{ nullptr }; // Lowest preceding operator node
        for (ParserToken& part : parts) {
            if (part.type() != ParserToken::Type::Operator)
                continue;

            if (part.order() > 0) // Order already set
                continue;

            processing = true;

            if (!lpn || ParserToken::operPrecedence(part.oper()) < ParserToken::operPrecedence(lpn->oper()))
                lpn = &part;
        }

        if (lpn)
            lpn->setOrder(++order);
    }


    /*
     * Calculate in order
    */
    processing = true;
    order = 1;
    while (processing) {
        processing = false;

        /*
         * Iterate over all operators.
         * Consider '1' '+' '2' '+' '3' - We will always want to start at the 2nd element and iterate every other.
        */
        for (auto partsIt = parts.begin() + 1; partsIt != parts.end(); partsIt += 2) {
            //auto& part = parts[i];
            ParserToken& oper = *partsIt;

            if (oper.order() == 0)
                throw ScriptException(EMSG::ExpectedOperator);

            if (oper.order() != order)
                continue;

            processing = true;

            try {
                auto& left = *(partsIt - 1);
                auto& right = *(partsIt + 1);
                partsIt = parts.emplace(partsIt + 2,
                                        oper.tokenBase(),
                                        calculate(left.value(), oper.oper(), right.value()));

            } catch (const ScriptException& ex) {
                throw ScriptException(oper.tokenBase().lineCol(), ex.m_originalMessage);
            }

            partsIt = parts.erase(partsIt - 3, partsIt);

            ++order;
            break;
        }
    }
    return parts.begin()->value();
}

std::vector<ValueHolder> Script::resolveParameters(const Scope &parameters, SymbolScope &symbols)
{
    std::vector<ValueHolder> ret;
    for (auto& paramex : parameters)
        ret.emplace_back(resolveExpression(paramex, symbols)); // TODO  consider std::transform ?
    return ret;
}

std::string Script::resolveCommandLine(const std::string& input, SymbolScope& symbols)
{
    std::string ret;
    enum {
        ST_Copy,
        ST_Variable
    } state = ST_Copy;
    std::string buf;
    for (auto i = 0u; i < input.size(); ++i) {
        const char c = input[i];
        if (state == ST_Copy) {
            if (c == '%') {
                state = ST_Variable;
                buf.clear();
                continue;
            }
            ret += c;
        }

        if (state == ST_Variable) {
            if (isspace(c) || c == '%' || i == input.size() - 1) {
                if (i == input.size() - 1 && c != '%')
                    buf += c;
                auto maybe = symbols.get(buf);
                if (maybe) {
                    Symbol& symbol = *maybe;
                    if (symbol.getType() == Symbol::Type::Value)
                        ret += symbol.getValue().toStdString();
                    if (isspace(c))
                        ret += c;
                }
                state = ST_Copy;
                continue;
            }
            buf += c;
        }
    }
    return ret;
}

void Script::resetError()
{
    m_lastError.clear();
    //m_lastErrorLinecol = {0, 0};
}
