/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#ifndef SCRIPT_H
#define SCRIPT_H

#include "SymbolScope.h"
#include "Tokenizer.h"

// This is IdealIRC specific. If you want to rip out the script engine,
// analyse the following header and resemble it to suit your needs.
#include "ScriptDialog/ScriptDialog.h"
// ==================

#include <string>
#include <optional>
#include <exception>
#include <functional>

using CommandHandler = std::function<void(const std::string&)>;
using DispatchHandler = std::function<void(const std::string&, std::vector<ValueHolder>&)>;

class Script
{
public:
    Script(const std::string& input, CommandHandler commandHandler_, DispatchHandler dispatchHandler_);
    Script(Script&& other);
    Script(const Script&) = delete;
    ~Script();

    Script& operator=(Script&) = delete;

    bool reload(const std::string& newInput);
    bool reload();
    std::optional<ValueHolder> runFunction(const std::string& identifier, std::vector<ValueHolder>& paramValues);
    bool runCommand(const std::string& input);
    bool dispatchEvent(const std::string& event, std::vector<ValueHolder>& paramValues);
    std::optional<ValueHolder> runScopeWithSymbols(const Scope& scope, SymbolScope& symbols);
    std::string lastError() const;
    bool hasError() const;

    const std::optional<ConstMenuRef> menu(const std::string& name) const;

    static void registerUniversalValue(const std::string& name, ValueHolder&& value);
    static void registerUniversalFunction(const std::string& name, const BuiltinFn& fnptr);

    int dialogConstruct(const std::string& name);
    ValueHolder dialogIsValidHandle(int handle);
    void dialogClose(int handle);
    void dialogHide(int handle);
    void dialogShow(int handle);
    ValueArray dialogList(const std::string& name);
    void dialogSetAttr(int handle, const std::string& attribute, ValueHolder& value);
    ValueHolder dialogGetAttr(int handle, const std::string& attribute);
    void dialogSetWidgetAttr(int handle, const std::string& widgetName, const std::string& attribute, ValueHolder& value);
    ValueHolder dialogGetWidgetAttr(int handle, const std::string& widgetName, const std::string& attribute);

    int dialogTableWidgetCount(int handle, const std::string& widgetName);
    void dialogTableWidgetInsert(int handle, const std::string& widgetName, const ValueArray& cols);
    void dialogTableWidgetRemove(int handle, const std::string& widgetName, int index);
    int dialogTableWidgetSelected(int handle, const std::string& widgetName);
    ValueArray dialogTableWidgetRow(int handle, const std::string& widgetName, int index);


private:
    static SymbolScope m_universal;
    std::optional<SymbolScope> m_glob;
    std::optional<SymbolScope> m_cmds;
    std::optional<SymbolScope> m_hooks;
    std::unordered_map<int,ScriptDialog> m_dlgs;
    int m_dlgHndlCount{ 0 };

    std::optional<Tokenizer> m_tokenizer;
    std::string m_data;

    CommandHandler commandHandler;
    DispatchHandler dispatchHandler;

    std::variant<std::optional<ValueHolder>, TokenBaseRef> runScope(const Scope &scope, SymbolScope& parentSymbols);
    ValueHolder resolveParamsRunScope(const Scope &scope, std::vector<ValueHolder>& paramValues, SymbolScope& symbols);
    ValueHolder resolveExpression(const Expression& expr, SymbolScope& symbols);
    std::vector<ValueHolder> resolveParameters(const Scope& parameters, SymbolScope& symbols);
    std::string resolveCommandLine(const std::string& input, SymbolScope& symbols);

    void resetError();
    std::string m_lastError;
    std::pair<int,int> m_lastErrorLinecol;
};

#endif // SCRIPT_H
