/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "SymbolScope.h"
#include "ScriptException.h"
#include <iostream>

Symbol::Symbol(const Scope& scope)
    : m_data(scope)
    , m_type(Type::Scope)
{}

Symbol::Symbol(const ValueHolder& data)
    : m_data(data)
    , m_type(Type::Value)
{}

Symbol::Symbol(const BuiltinFn& fn)
    : m_data(fn)
    , m_type(Type::Builtin)
{}

Symbol::Type Symbol::getType() const
{
    return m_type;
}

ValueHolder& Symbol::getValue()
{
    return std::get<ValueHolder>(m_data);
}

const ValueHolder& Symbol::getValue() const
{
    return std::get<ValueHolder>(m_data);
}

const Scope& Symbol::getScope() const
{
    return std::get<ScopeCRef>(m_data);
}

ValueHolder Symbol::runBuiltin(Script& script, std::vector<ValueHolder>& params)
{
    if (m_type != Type::Builtin)
        return {};
    auto fn = std::get<BuiltinFn>(m_data);
    return fn(script, params);
}

SymbolScope::SymbolScope(SymbolScope* parent)
    : m_parent(parent)
{}

SymbolScope::SymbolScope(SymbolScope&& other)
    : m_symbols(std::move(other.m_symbols))
    , m_parent(other.m_parent)
{}

void SymbolScope::setParent(SymbolScope& parent)
{
    m_parent = &parent;
}

void SymbolScope::set(const std::string& name, const ValueHolder& value)
{
    auto existing = get(name);
    if (existing)
        existing->get() = value;
    else
        m_symbols.insert_or_assign(name, value);
}

void SymbolScope::set(const std::string& name, ValueHolder&& value)
{
    auto existing = get(name);
    if (existing)
        existing->get() = value;
    else
        m_symbols.insert_or_assign(name, std::move(value));
}

void SymbolScope::set(const std::string& name, Scope& scope)
{
    m_symbols.insert_or_assign(name, scope);
}

void SymbolScope::set(const std::string& name, const BuiltinFn& fn)
{
    m_symbols.insert_or_assign(name, fn);
}

std::optional<SymbolRef> SymbolScope::get(const std::string& name)
{
    auto iter = m_symbols.find(name);
    if (iter == m_symbols.end())
        return m_parent ? m_parent->get(name) : std::nullopt;
    else
        return m_symbols.at(name);
}

std::optional<SymbolCRef> SymbolScope::cget(const std::string& name) const
{
    auto iter = m_symbols.find(name);
    if (iter == m_symbols.end())
        return m_parent ? m_parent->cget(name) : std::nullopt;
    else
        return m_symbols.at(name);
}

void SymbolScope::clear()
{
    m_symbols.clear();
}

void SymbolScope::dump(bool deep) const
{
    std::cout << "- - - - -\n";
    for (auto& [key, val] : m_symbols) {
        std::cout << key << " = ";
        if (val.getType() == Symbol::Type::Scope)
            std::cout << "[Scope]\n";
        else
            std::cout << "'" << val.getValue().toStdString() << "'\n";
    }

    if (deep && m_parent)
            m_parent->dump(true);
    else
        std::cout << "- - - - -\n";
}

bool SymbolScope::hasValues() const
{
    return !m_symbols.empty();
}
