/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#ifndef TOKENS_H
#define TOKENS_H

#include <string>
#include <vector>
#include <memory>

enum class STokenType {
    Undefined,
    Command,
    Function,
    Keyword,
    MapKeyword,
    Reference,
    BoolLiteral,
    Hook,
    Dispatch,

    Operator,
    StringLiteral,
    NumberLiteral,
    RealLiteral,
    Comma,
    SemiColon,

    CondIf,
    CondElse,
    CondWhile,

    Continue,
    Break,

    Scope, // Triggers at opening curly-parenthesis, signal a scope of expressions / statements.
    Expression, // Triggers at opening parenthesis, with an operator before, signal a sub-expression.
    Parameters, // Triggers at opening parenthesis, with a keyword before, signals a list of function parameters

    /* Not actual token types */
    EndScope, // Used to signal end of a scope of expressions / statements
    EndParameters, // Used to signal end of function parameter list
    EndExpression, // Used to signal end of a sub-expression (parenthesises)
    Comment // Used during tokenization for stripping comments.
};

enum class ImperativeBlockType {
    Function,
    Command,
    Hook
};

/* ************************************************************** */
/* ************************************************************** */

template<typename T>
class VectorOfUniquePtrsIterator
{
    typename std::vector<std::unique_ptr<T>>::iterator it;

public:
    VectorOfUniquePtrsIterator() = delete;
    ~VectorOfUniquePtrsIterator() = default;
    VectorOfUniquePtrsIterator(typename std::vector<std::unique_ptr<T>>::iterator underlyingIter)
        : it(underlyingIter)
    {}

    VectorOfUniquePtrsIterator(const VectorOfUniquePtrsIterator& other)
        : it(other.it)
    {}

    VectorOfUniquePtrsIterator(VectorOfUniquePtrsIterator&& other)
        : it(std::move(other.it))
    {}

    VectorOfUniquePtrsIterator& operator++()
    {
        ++it;
        return *this;
    }

    VectorOfUniquePtrsIterator& operator--()
    {
        --it;
        return *this;
    }

    VectorOfUniquePtrsIterator operator+(int i)
    {
        return it + i;
    }

    VectorOfUniquePtrsIterator operator-(int i)
    {
        return it - i;
    }

    bool operator==(const VectorOfUniquePtrsIterator& other)
    {
        return it == other.it;
    }

    bool operator!=(const VectorOfUniquePtrsIterator& other)
    {
        return it != other.it;
    }

    T& operator*()
    {
        return **it;
    }

    T* operator->()
    {
        return &(**it);
    }
};

/* ************************************************************** */
/* ************************************************************** */

class Expression;
class Scope;
class ConditionalScope;

/* ************************************************************** */
/* ************************************************************** */

class TokenBase
{
public:
    virtual ~TokenBase() = default;
    STokenType type() const;
    std::pair<int,int> lineCol() const;

protected:
    TokenBase(TokenBase&& other);
    explicit TokenBase(STokenType type, int line, int col);

private:
    const STokenType m_type;
    const int m_line;
    const int m_col;
};

using TokenBaseRef = std::reference_wrapper<TokenBase>;

/* ************************************************************** */
/* ************************************************************** */

class Token : public TokenBase
{
    friend class Expression;
    friend class MapToken;
    friend class DispatchToken;

public:
    ~Token() override = default;
    const std::string& token() const;

private:
    Token(const std::string& token, STokenType type, int line, int col);
    Token(Token&& other);

    const std::string m_token;
};

/* ************************************************************** */
/* ************************************************************** */

class BoolToken : public TokenBase
{
    friend class Expression;
    friend class MapToken;
    friend class DispatchToken;

public:
    ~BoolToken() override = default;
    bool value() const;

private:
    BoolToken(bool val, int line, int col);
    BoolToken(BoolToken&& other);

    const bool m_val;
};

/* ************************************************************** */
/* ************************************************************** */

struct MapTokenPriv;
class MapToken : public Token
{
public:
    MapToken(const std::string& token, int line, int col, bool isRef = false);
    MapToken(MapToken&& other);
    ~MapToken() override;

    Expression& subscript();
    const Expression& subscript() const;

private:
    std::unique_ptr<MapTokenPriv> mp;
};

/* ************************************************************** */
/* ************************************************************** */

struct ListConstructPriv;
class ListConstruct : public Token // TODO ListConstruct is not yet fully implemented.
{
public:
    ListConstruct(int line, int col);
    ListConstruct(ListConstruct&& other);
    ~ListConstruct() override;

    const std::vector<Expression>& items();
    Expression& addItem(int line, int col);

private:
    std::unique_ptr<ListConstructPriv> mp;
};

/* ************************************************************** */
/* ************************************************************** */

struct DispatchTokenPriv;
class DispatchToken : public Token
{
public:
    DispatchToken(const std::string& token, int line, int col);
    DispatchToken(DispatchToken&& other);
    ~DispatchToken() override;

    Scope& parameters();

private:
    std::unique_ptr<DispatchTokenPriv> mp;
};

/* ************************************************************** */
/* ************************************************************** */

struct FunctionTokenPriv;
class FunctionToken : public TokenBase
{
    friend class Expression;

public:
    ~FunctionToken() override;
    Scope& parameters();
    const Scope& parameters() const;
    const std::string& identifier() const;

private:
    FunctionToken(const std::string& identifier, int line, int col);
    FunctionToken(FunctionToken&& other);

    std::unique_ptr<FunctionTokenPriv> mp;
};

/* ************************************************************** */
/* ************************************************************** */

struct ConditionalScopeTokenPriv;
class ConditionalScopeToken : public TokenBase
{
    friend class Expression;

public:
    ~ConditionalScopeToken() override;
    ConditionalScope& scope() const;

private:
    ConditionalScopeToken(STokenType type, int line, int col, Expression&& condition);
    std::unique_ptr<ConditionalScopeTokenPriv> mp;
};

/* ************************************************************** */
/* ************************************************************** */

struct ExpressionPriv;
class Expression : public TokenBase
{
public:
    Expression(int line, int col);
    Expression(Expression&& other);
    Expression(Expression&) = delete;
    ~Expression() override;

    using Iterator = VectorOfUniquePtrsIterator<TokenBase>;

    Expression& operator=(Expression&) = delete;
    Expression& operator=(Expression&&);

    void addToken(const std::string& token, STokenType type, int line, int col);
    void addBool(bool val, int line, int col);
    MapToken& addMapToken(const std::string& token, int line, int col, bool isRef = false);
    DispatchToken& addDispatchToken(const std::string& token, int line, int col);
    Expression& addExpression(int line, int col);
    FunctionToken& addFunction(const std::string& identifier, int line, int col);
    ConditionalScopeToken& addIfScope(int line, int col, Expression&& condition);
    ConditionalScopeToken& addElseScope(int line, int col);
    ConditionalScopeToken& addWhileScope(int line, int col, Expression&& condition);

    Iterator begin() const;
    Iterator end() const;
    size_t count() const;
    void eraseBack();

    TokenBase& operator[](size_t idx) const;
    TokenBase& back() const;

private:
    std::unique_ptr<ExpressionPriv> mp;
};

/* ************************************************************** */
/* ************************************************************** */

struct ScopePriv;
using ScopeRef = std::reference_wrapper<Scope>;
class Scope
{
public:
    Scope(int line, int col);
    explicit Scope(std::pair<int,int> lineCol);
    Scope(Scope&& other);
    Scope(Scope&) = delete;
    virtual ~Scope();

    std::pair<int,int> getLineCol() const;
    Expression& addStatement(Expression&& expr);
    std::size_t countStatements() const;

    std::vector<Expression>::iterator begin() const;
    std::vector<Expression>::iterator end() const;

private:
    std::unique_ptr<ScopePriv> mp;
    const int m_line;
    const int m_col;
};

/* ************************************************************** */
/* ************************************************************** */

class ImperativeScope : public Scope
{
public:
    ImperativeScope(int line, int col, ImperativeBlockType type, const std::string& identifier, Scope&& params);
    ~ImperativeScope() override = default;
    const std::string& identifier() const;
    const Scope& params() const;
    ImperativeBlockType type() const;

private:
    ImperativeBlockType m_type;
    const std::string m_id;
    Scope m_params;
};

class ConditionalScope : public Scope
{
public:
    ConditionalScope(int line, int col, Expression&& condition);
    ~ConditionalScope() override = default;
    const Expression& condition() const;

private:
    Expression m_cond;
};

#endif // TOKENS_H
