/*
 * IdealIRC Script Engine - Scripting tailored for IRC clients.
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#ifndef VALUEHOLDER_H
#define VALUEHOLDER_H

#include "ParserOperator.h"
#include <optional>
#include <variant>
#include <string>
#include <memory>
#include <unordered_map>

class ValueHolder;
using ValueHolderPtr = std::unique_ptr<ValueHolder>;
using ValueArray = std::unordered_map<std::string, ValueHolderPtr>;
using ValueRef = std::reference_wrapper<ValueHolder>;

class ValueHolder
{
    friend class ValueExtract;
    using ValueTypes = std::optional< std::variant<ValueRef, int, double, bool, std::string, ValueArray> >;
    ValueTypes m_val;
    bool m_const{ false };

public:
    enum class Type {
        Void,
        Integer,
        Real,
        Bool,
        String,
        Map // Here's an anti-pattern for ya: Please always keep this the last entry of this enum.
    };

    ValueHolder() = default;

    ValueHolder(const ValueHolder& other);

    ValueHolder(ValueHolder&& other);

    ~ValueHolder();

    ValueHolder(int val);

    ValueHolder(double val);

    ValueHolder(bool val);

    ValueHolder(const std::string& val);

    ValueHolder(const ValueArray& array);

    ValueHolder(ValueArray&& array);

    ValueHolder(const ValueRef& ref);

    ValueHolder& operator=(int val);

    ValueHolder& operator=(double val);

    ValueHolder& operator=(bool val);

    ValueHolder& operator=(const std::string& val);

    ValueHolder& operator=(const ValueHolder& other);

    ValueHolder& operator=(const ValueArray& val);

    ValueHolder& operator[](const std::string& subscript);

    const ValueHolder& operator[](const std::string& subscript) const;


    bool rangedOperator(ParserOperator op, const ValueHolder& other) const;

    bool equalityOperator(ParserOperator op, const ValueHolder& other) const;

    ValueHolder arithmeticOperator(ParserOperator op, const ValueHolder& other) const;

    bool logicOperator(ParserOperator op, const ValueHolder& other) const;


    bool operator<(const ValueHolder& other) const;

    bool operator>(const ValueHolder& other) const;

    bool operator<=(const ValueHolder& other) const;

    bool operator>=(const ValueHolder& other) const;

    bool operator==(const ValueHolder& other) const;

    bool operator!=(const ValueHolder& other) const;

    ValueHolder operator+(const ValueHolder& other) const;

    ValueHolder operator-(const ValueHolder& other) const;

    ValueHolder operator*(const ValueHolder& other) const;

    ValueHolder operator/(const ValueHolder& other) const;

    ValueHolder operator%(const ValueHolder& other) const;

    bool operator&&(const ValueHolder& other) const;

    bool operator||(const ValueHolder& other) const;


    void clear(); //!< Clear the value by following references, or self if not a reference.

    void reset(); //!< Clear the value including reference

    bool hasValue() const;

    Type getType() const;

    std::string toStdString() const;

    void setArray(const std::string& subscript, const ValueHolder& val);

    bool isReference() const;

    void markAsConst();

    bool isConst() const;

private:
    ValueTypes& mval();
    const ValueTypes& mval() const;
};

#endif // VALUEHOLDER_H
