/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#ifndef IIRCVIEW_H
#define IIRCVIEW_H

#include <QTextBrowser>
#include <QResizeEvent>
#include <QTextFrame>
#include <QMouseEvent>

enum class PrintType {
    Normal = 0, //!< Normal text
    ProgramInfo, //!< Information text from IdealIRC
    ServerInfo, //!< Uncategorized messages from the server
    Nick, //!< Nickname changes
    Mode, //!< Mode setting on local user and channel
    Quit, //!< Quit messages
    Join, //!< Channel join messages
    Part, //!< Channel part messages
    Topic, //!< Topic messages (covers also the numerics)
    Invite, //!< Invitation to a channel
    Kick, //!< Channel kick messages
    Wallops, //!< Wallop messages
    Notice, //!< Notice messages
    Action, //!< Action/Role-play styled message ("/me does a thing")
    CTCP, //!< CTCP request and reply
    OwnText, //!< Messages sent from us
    Highlight //!< Messages marked as highlighted
};

enum class ColorCode {
    White,
    Black,
    Blue,
    Green,
    BrightRed,
    Red,
    Magenta,
    Brown,
    Yellow,
    Brightgreen,
    Cyan,
    BrightCyan,
    BrightBlue,
    BrightMagenta,
    DarkGray,
    LightGray
};

QColor colorCode(ColorCode cc);

enum CtrlCode : char {
    CtrlBold			= 0x02,
    CtrlUnderline		= 0x1F,
    CtrlColor			= 0x03,
    CtrlReverse			= 0x16,
    CtrlStriketrhough	= 0x13,
    CtrlReset			= 0x0F
};

bool isCtrlCode(char c);

class IIRCView : public QTextBrowser
{
    Q_OBJECT

public:
    IIRCView();
    void print(const QDateTime& timestamp, PrintType ptype, const QString& text);
    void resetView();
    static QString stripTags(const QString& text);
    static QString formatType(PrintType ptype, const QString& text);
    static QString typeToString(PrintType ptype, const QString& defaultValue = "");

signals:
    void customContextMenuRequested(const QPoint& point);

private:
    QTextFrame* topMargin{ nullptr };
    void resizeEvent(QResizeEvent* evt) override;
    void mouseReleaseEvent(QMouseEvent* evt) override;
    int reduceMargin() const;
    void newConfiguration();
    static void formatSpaces(QString& text);
    static void ctrlFormat(QString& text);
    static void linkFormat(QString& text);

    /** @brief
     Since Qt is a good boy and formats the "HTML" by inlining the "CSS" styles right in the tags,
     we need to cache the text we append as "HTML", in order to re-set colors upon
     style re-definitions (via configuration).
    */
    QString cachedText;
};

#endif // IIRCVIEW_H
