/*
 * IdealIRC - Internet Relay Chat client
 * Copyright (C) 2021  Tom-Andre Barstad.
 * This software is licensed under the Software Attribution License.
 * See LICENSE for more information.
*/

#include "ILineEdit.h"
#include "ConfigMgr.h"
#include <QMessageBox>
#include <IWin/IWin.h>

enum CtrlModifierCode : uint16_t {
    CMC_Reverse          = 0x25E9,
    CMC_Bold            = 0x25A0,
    CMC_Underline       = 0x25BD,
    CMC_Color           = 0x25A9,
    CMC_Strikethrough   = 0x25BB,
    CMC_Reset           = 0x25A1
};

ILineEdit::ILineEdit(ILineEdit::TabCompleteCb&& tabCompleteCb, QWidget* parent)
    : QLineEdit(parent)
    , tabCb(std::move(tabCompleteCb))
{
    ConfigMgr& conf = ConfigMgr::instance();
    connect(&conf, &ConfigMgr::saved, this, &ILineEdit::loadConfig);
    loadConfig();

    auto* filter = new KeyEventFilter(this);
    connect(filter, &KeyEventFilter::newLine,
        [this](const QString& l) {
            emit newLine(l);
            clear();
        });
    installEventFilter(filter);
}

void ILineEdit::loadConfig()
{
    ConfigMgr& conf = ConfigMgr::instance();

    const QString style = QStringLiteral("QLineEdit { "
                            "background-color: %1;"
                            "color: %2;"
                            "}")
                            .arg(conf.color("InputBackground"))
                            .arg(conf.color("InputForeground"));
    setStyleSheet(style);
    QFont newfont(conf.common("Font"));
    newfont.setPixelSize(conf.common("FontSize").toInt());
    setFont(newfont);
}

void ILineEdit::keyTab()
{
    const auto stxt = text().toStdString();
    const auto cpos = cursorPosition() - 1;

    if (cpos < 0)
        return;

    const int beginpos = [&stxt, &cpos] {
        int i = cpos;
        for (; i > -1; --i) {
            if (std::isspace(stxt[i]))
                break;
        }
        return i + 1;
    }();

    const int tokenlen = cpos - beginpos + 1;

    if (completePattern.isEmpty()) {
        completePattern = text().mid(beginpos, tokenlen);
    }

    if (!completePattern.isEmpty()) {
        auto patt = completePattern;
        int cidx = tabCb(completeidx, patt);

        if (patt != completePattern) {
            completeidx = cidx;
            auto ctext = text();
            ctext.replace(beginpos, tokenlen, patt);
            setText(ctext);
            setCursorPosition(beginpos + patt.length());
        }
    }
}

void ILineEdit::keyUp()
{
    if (logidx == -1)
        currentLine = text();
    if (logidx+1 >= prevLines.size())
        return;

    ++logidx;
    setText(prevLines[logidx]);
}

void ILineEdit::keyDown()
{
    if (logidx == -1)
        return;

    --logidx;
    if (logidx == -1) {
        setText(currentLine);
        return;
    }
    else
        setText(prevLines[logidx]);
}

void ILineEdit::resetCompletePattern()
{
    completePattern.clear();
    completeidx = 0;
}

void ILineEdit::keyPressEvent(QKeyEvent* evt)
{
    QLineEdit::keyPressEvent(evt);
}


bool KeyEventFilter::eventFilter(QObject* obj, QEvent* evt)
{

    auto* input = qobject_cast<ILineEdit*>(obj);
    if (!input)
        qWarning() << "Expected parent for KeyEventFilter is not an ILineEdit!";
    else if (evt->type() == QEvent::KeyPress) {
        auto kevt = dynamic_cast<QKeyEvent*>(evt);
        const bool Ctrl = kevt->modifiers() & Qt::ControlModifier;

        if (kevt->key() != Qt::Key_Tab)
            input->resetCompletePattern();

        if (kevt->key() == Qt::Key_Tab) {
            input->keyTab();
            return true;
        }
        else if (kevt->key() == Qt::Key_Up) {
            input->keyUp();
            return true;
        }
        else if (kevt->key() == Qt::Key_Down) {
            input->keyDown();
            return true;
        }
        else if (kevt->key() == Qt::Key_Return || kevt->key() == Qt::Key_Enter) {
            auto ln = input->text();
            ln.remove('\r');
            auto linelist = ln.split('\n');

            bool proceedPasting{ true };

            const auto* parent = dynamic_cast<IWin*>(input->parentWidget());

            if (linelist.count() > 3 && parent &&
            (parent->getType() == IWin::Type::Channel || parent->getType() == IWin::Type::Private))
            {
                auto msg = tr("You are about to paste %1 lines.\nYou should consider using an external \"paste service\".\nContinue anyway?")
                                .arg(linelist.count());
                auto btn = QMessageBox::warning(input, tr("Flood warning"), msg, QMessageBox::Yes | QMessageBox::No);
                proceedPasting = btn == QMessageBox::Yes;
            }
            if (proceedPasting) {
                for (auto line : linelist) {
                    input->prevLines.push_front(line);
                    if (!Ctrl) {
                        line.replace(QChar(CMC_Reverse), QChar(CtrlReverse));
                        line.replace(QChar(CMC_Bold), QChar(CtrlBold));
                        line.replace(QChar(CMC_Underline), QChar(CtrlUnderline));
                        line.replace(QChar(CMC_Color), QChar(CtrlColor));
                        line.replace(QChar(CMC_Strikethrough), QChar(CtrlStriketrhough));
                        line.replace(QChar(CMC_Reset), QChar(CtrlReset));
                    }
                    emit newLine(line);
                }
            }
        }
        else if (Ctrl && kevt->key() == Qt::Key_I) {
            input->insert(QChar(CMC_Reverse));
            return true;
        }
        else if (Ctrl && kevt->key() == Qt::Key_B) {
            input->insert(QChar(CMC_Bold));
            return true;
        }
        else if (Ctrl && kevt->key() == Qt::Key_U) {
            input->insert(QChar(CMC_Underline));
            return true;
        }
        else if (Ctrl && kevt->key() == Qt::Key_K) {
            input->insert(QChar(CMC_Color));
            return true;
        }
        else if (Ctrl && kevt->key() == Qt::Key_S) {
            input->insert(QChar(CMC_Strikethrough));
            return true;
        }
        else if (Ctrl && kevt->key() == Qt::Key_R) {
            input->insert(QChar(CMC_Reset));
            return true;
        }
        else {
            input->logidx = -1;
        }
    }
    return QObject::eventFilter(obj, evt);
}
